"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import numpy
import numpy.lib.recfunctions
import scipy.optimize
import pandas
import gemmi
from servalcat.utils import logger

dtypes64 = dict(i=numpy.int64, u=numpy.uint64, f=numpy.float64, c=numpy.complex128)
to64 = lambda x: x.astype(dtypes64.get(x.dtype.kind, x.dtype))

class Binner:
    def __init__(self, asu, style="relion"):
        if style == "relion":
            cell = asu.unit_cell
            max_cell_edge = numpy.max([cell.a, cell.b, cell.c])
            self.d_array = asu.make_d_array()
            self.bin_array = (max_cell_edge/self.d_array).astype(int)
            self.bins, self.bin_counts = numpy.unique(self.bin_array, return_counts=True)
        else:
            raise Exception("Non-supported binning type")
    # __init__()
# class Binner    

def r_factor(fo, fc):
    return numpy.nansum(numpy.abs(fo-fc)) / numpy.nansum(fo)

def df_from_asu_data(asu_data, label):
    df = pandas.DataFrame(data=asu_data.miller_array,
                          columns=["H","K","L"])
    if asu_data.value_array.dtype.names == ('value', 'sigma'):
        df[label] = to64(asu_data.value_array["value"])
        df["SIG"+label] = to64(asu_data.value_array["sigma"])
    else:
        df[label] = to64(asu_data.value_array)
    return df

def df_from_raw(miller_array, value_array, label):
    df = pandas.DataFrame(data=miller_array,
                          columns=["H","K","L"])
    df[label] = to64(value_array)
    return df

def hkldata_from_asu_data(asu_data, label):
    df = df_from_asu_data(asu_data, label)
    return HklData(asu_data.unit_cell, asu_data.spacegroup, df)
# hkldata_from_asu_data()

def hkldata_from_mtz(mtz, labels, newlabels=None):
    assert type(mtz) == gemmi.Mtz
    if not set(labels).issubset(mtz.column_labels()):
        raise RuntimeError("All specified coulumns were not found from mtz.")
    
    col_types = {x.label:x.type for x in mtz.columns}
    df = pandas.DataFrame(data=numpy.array(mtz, copy=False), columns=mtz.column_labels())
    df = df.astype({col: 'int32' for col in col_types if col_types[col] == "H"})
    df = df.astype({col: 'Int64' for col in col_types if col_types[col] in ("B", "Y", "I")}) # pandas's nullable int
    for lab in set(mtz.column_labels()).difference(labels+["H","K","L"]):
        del df[lab]
        
    if newlabels is not None:
        assert len(newlabels) == len(labels)
        for i in range(1, len(newlabels)):
            if newlabels[i] == "": # means this is phase and should be transferred to previous column
                assert col_types.get(labels[i]) == "P"
                assert col_types.get(labels[i-1]) == "F"
                ph = numpy.deg2rad(df[labels[i]])
                df[labels[i-1]] = df[labels[i-1]] * (numpy.cos(ph) + 1j * numpy.sin(ph))
                del df[labels[i]]
        
        df.rename(columns={x:y for x,y in zip(labels, newlabels) if y != ""}, inplace=True)

    return HklData(mtz.cell, mtz.spacegroup, df)
# hkldata_from_mtz()

def blur_mtz(mtz, B):
    # modify given mtz object
    
    s2 = mtz.make_1_d2_array()
    k2 = numpy.exp(-B*s2/2)
    k = numpy.exp(-B*s2/4)
    i_labs = [c.label for c in mtz.columns if c.type in "JK"]
    f_labs = [c.label for c in mtz.columns if c.type in "FDG"]
    for labs in i_labs, f_labs:
        for l in labs:
            sl = "SIG"+l
            if sl in mtz.column_labels(): labs.append(sl)

    if i_labs:
        logger.writeln("Intensities: {}".format(" ".join(i_labs)))
        logger.writeln("  exp(-B*s^2/2) will be multiplied (B= {:.2f})".format(B))
    if f_labs:
        logger.writeln("Amplitudes:  {}".format(" ".join(f_labs)))
        logger.writeln("  exp(-B*s^2/4) will be multiplied (B= {:.2f})".format(B))

    for l in i_labs:
        c = mtz.column_with_label(l)
        c.array[:] *= k2
    for l in f_labs:
        c = mtz.column_with_label(l)
        c.array[:] *= k
# blur_mtz()

def mtz_selected(mtz, columns):
    """
    creates a new mtz object having specified `columns` of `mtz`
    """
    columns = ["H", "K", "L"] + columns # TODO make sure no duplicates
    col_dict = {x.label:x for x in mtz.columns}
    col_idxes = {x.label:i for i, x in enumerate(mtz.columns)}

    notfound = list(set(columns) - set(col_idxes))
    if notfound:
        raise RuntimeError("specified columns not found: {}".format(str(notfound)))

    # copy metadata
    mtz2 = gemmi.Mtz()
    for k in ("spacegroup", "cell", "history", "title"):
        setattr(mtz2, k, getattr(mtz, k))

    for ds in mtz.datasets:
        ds2 = mtz2.add_dataset("")
        for k in ("cell", "id", "crystal_name", "dataset_name", "project_name", "wavelength"):
            setattr(ds2, k, getattr(ds, k))

    # copy selected columns
    for col in columns:
        mtz2.add_column(col, col_dict[col].type,
                        dataset_id=col_dict[col].dataset_id, expand_data=False)

    idxes = [col_idxes[col] for col in columns]
    data = numpy.array(mtz, copy=False)[:, idxes]
    mtz2.set_data(data)
    return mtz2
# mtz_selected()
    
class HklData:
    def __init__(self, cell, sg, df=None, binned_df=None):
        self.cell = cell
        self.sg = sg
        self.df = df
        self.binned_df = binned_df
        self._bin_and_indices = []
    # __init__()

    def update_cell(self, cell):
        # update d
        pass

    def switch_to_asu(self):
        # Need to care phases
        assert not any(numpy.iscomplexobj(self.df[x]) for x in self.df)
        hkl = self.miller_array().to_numpy()
        self.sg.switch_to_asu(hkl)
        self.df[["H","K","L"]] = hkl

    def copy(self, d_min=None, d_max=None):
        # FIXME we should reset_index here? after resolution truncation, max(df.index) will be larger than size.
        if (d_min, d_max).count(None) == 2:
            df = self.df.copy()
            binned_df = self.binned_df.copy() if self.binned_df is not None else None
        else:
            if d_min is None: d_min = 0
            if d_max is None: d_max = float("inf")
            d = self.d_spacings()
            sel = (d >= d_min) & (d <= d_max)
            df = self.df[sel].copy()
            binned_df = None # no way to keep it
        
        return HklData(self.cell, self.sg,  df, binned_df)
    # copy()

    def merge_asu_data(self, asu_data, label, common_only=True):
        if self.df is not None and label in self.df:
            raise Exception("Duplicated label")
        
        df_tmp = df_from_asu_data(asu_data, label)

        if self.df is None:
            self.df = df_tmp
        elif common_only:
            self.df = self.df.merge(df_tmp)
        else:
            self.df = self.df.merge(df_tmp, how="outer")
    # merge_asu_data()

    def miller_array(self):
        return self.df[["H","K","L"]]

    def s_array(self):
        hkl = self.miller_array()
        return numpy.dot(hkl, self.cell.fractionalization_matrix)

    def debye_waller_factors(self, b_cart=None, b_iso=None):
        if b_iso is not None:
            s2 = 1 / self.d_spacings()**2
            return numpy.exp(-b_iso / 4 * s2)
        if b_cart is not None:
            b_star = b_cart.transformed_by(self.cell.fractionalization_matrix)
            return numpy.exp(-b_star.r_u_r(self.miller_array().to_numpy()) / 4)
    
    def calc_d(self):
        self.df["d"] = self.cell.calculate_d_array(self.miller_array())
    # calc_d()

    def calc_epsilon(self):
        self.df["epsilon"] = self.sg.operations().epsilon_factor_without_centering_array(self.miller_array())
    # calc_epsilon()

    def calc_centric(self):
        self.df["centric"] = self.sg.operations().centric_flag_array(self.miller_array()).astype(int)
    # calc_centric()
        
    def d_spacings(self):
        if "d" not in self.df or self.df.d.isnull().values.any():
            self.calc_d()
        return self.df.d
    # calc_d()

    def sort_by_resolution(self, ascending=False):
        self.d_spacings()
        self.df.sort_values("d", ascending=ascending, inplace=True)
    # sort_by_resolution()

    def d_min_max(self):
        d = self.d_spacings()
        return numpy.min(d), numpy.max(d)
    # d_min_max()

    def complete(self):
        # make complete set
        all_hkl = gemmi.make_miller_array(self.cell, self.sg, self.d_min_max()[0])
        match = gemmi.HklMatch(self.miller_array(), all_hkl)
        missing_hkl_df = pandas.DataFrame(all_hkl[numpy.asarray(match.pos) < 0], columns=["H","K","L"])
        self.df = pandas.concat([self.df, missing_hkl_df])
        logger.writeln("Completing hkldata: {} reflections were missing".format(len(missing_hkl_df.index)))
        self.calc_d()
    # complete()

    def completeness(self, label=None):
        if label is None:
            n_missing = numpy.sum(self.df.isna().any(axis=1))
        else:
            n_missing = numpy.sum(self.df[label].isna())
        n_all = len(self.df.index)
        return (n_all-n_missing)/n_all
    # completeness()

    def setup_binning(self, n_bins, method=gemmi.Binner.Method.Dstar2):
        self.df.reset_index(drop=True, inplace=True)
        s2 = 1/self.d_spacings().to_numpy()**2
        binner = gemmi.Binner()
        binner.setup_from_1_d2(n_bins, method, s2, self.cell)
        self._bin_and_indices = []
        d_limits = 1 / numpy.sqrt(binner.limits)
        bin_number = binner.get_bins_from_1_d2(s2)
        d_max_all = []
        d_min_all = []
        for i in range(binner.size):
            left = numpy.max(self.d_spacings()) if i == 0 else d_limits[i-1]
            right = numpy.min(self.d_spacings()) if i == binner.size -1 else d_limits[i]
            sel = numpy.where(bin_number==i)[0] # slow?
            d_max_all.append(left)
            d_min_all.append(right)
            self._bin_and_indices.append((i, sel))

        self.df["bin"] = bin_number
        self.binned_df = pandas.DataFrame(dict(d_max=d_max_all, d_min=d_min_all), index=list(range(binner.size)))
    # setup_binning()

    def setup_relion_binning(self, sort=False):
        max_edge = max(self.cell.parameters[:3])
        if sort:
            self.sort_by_resolution()
        self.df.reset_index(drop=True, inplace=True) # to allow numpy.array indexing
            
        self.df["bin"] = (max_edge/self.d_spacings()+0.5).astype(int)
        # Merge inner/outer shells if too few # TODO smarter way
        bin_counts = []
        bin_ranges = {}
        modify_table = {}
        for i_bin, g in self.df.groupby("bin", sort=True):
            if i_bin == 0: continue # ignore DC component
            bin_counts.append([i_bin, g.index])
            bin_ranges[i_bin] = (numpy.max(g.d), numpy.min(g.d))

        for i in range(len(bin_counts)):
            if len(bin_counts[i][1]) < 10 and i < len(bin_counts)-1:
                bin_counts[i+1][1] = bin_counts[i+1][1].union(bin_counts[i][1])
                modify_table[bin_counts[i][0]] = bin_counts[i+1][0]
                logger.writeln("Bin {} only has {} data. Merging with next bin.".format(bin_counts[i][0],
                                                                                      len(bin_counts[i][1])))
            else: break

        for i in reversed(range(len(bin_counts))):
            if i > 0 and len(bin_counts[i][1])/len(bin_counts[i-1][1]) < 0.5:
                bin_counts[i-1][1] = bin_counts[i-1][1].union(bin_counts[i][1])
                modify_table[bin_counts[i][0]] = bin_counts[i-1][0]
                logger.writeln("Bin {} only has {} data. Merging with previous bin.".format(bin_counts[i][0],
                                                                                          len(bin_counts[i][1])))
            else: break

        while True:
            flag = True
            for i_bin in modify_table:
                if modify_table[i_bin] in modify_table:
                    modify_table[i_bin] = modify_table[modify_table[i_bin]]
                    flag = False
            if flag: break

        for i_bin in modify_table:
            new_bin = modify_table[i_bin]
            self.df["bin"] = numpy.where(self.df["bin"].to_numpy() == i_bin, new_bin, self.df["bin"].to_numpy())
            bin_ranges[new_bin] = (max(bin_ranges[i_bin][0], bin_ranges[new_bin][0]),
                                   min(bin_ranges[i_bin][1], bin_ranges[new_bin][1]))

        self._bin_and_indices = []
        bin_all = []
        d_max_all = []
        d_min_all = []
        for i_bin, indices in bin_counts:
            if i_bin in modify_table: continue
            #if sort: # want this, but we cannot take len() for slice. we can add ncoeffs to binned_df
            #    self._bin_and_indices.append((i_bin, slice(numpy.min(indices), numpy.max(indices))))
            #else:
            self._bin_and_indices.append((i_bin, indices))
                
            bin_all.append(i_bin)
            d_max_all.append(bin_ranges[i_bin][0])
            d_min_all.append(bin_ranges[i_bin][1])
        self.binned_df = pandas.DataFrame(dict(d_max=d_max_all, d_min=d_min_all), index=bin_all)
    # setup_relion_binning()

    def binned_data_as_array(self, lab):
        vals = numpy.zeros(len(self.df.index), dtype=self.binned_df[lab].dtype)
        for i_bin, idxes in self.binned():
            vals[idxes] = self.binned_df[lab][i_bin]
        return vals
    # binned_data_as_array()

    def binned(self):
        return self._bin_and_indices
    
    def merge(self, other, common_only=True):
        self.merge_df(other, common_only)
    # merge()

    def merge_df(self, other, common_only=True):
        # TODO check space group, cell
        # TODO transform to asu with phase shifts
        # TODO check column labels. same names other than HKL?
        # 
        if common_only:
            self.df = self.df.merge(other)
        else:
            df = self.df.merge(other, indicator=True, how="outer")
            df_left = df[df._merge=="left_only"]
            df_right = df[df._merge=="right_only"]
            df_both = df[df._merge=="both"]
    # merge()

    def as_numpy_arrays(self, labels, omit_nan=True):
        tmp = self.df[labels]
        if omit_nan: tmp = tmp[~tmp.isna().any(axis=1)]
        return [tmp[lab].to_numpy() for lab in labels]
    # as_numpy_arrays()

    def as_asu_data(self, label=None, data=None, label_sigma=None):
        if label is None: assert data is not None
        else: assert data is None

        if label_sigma is not None:
            assert data is None
            assert not numpy.iscomplexobj(self.df[label])
            sigma = self.df[label_sigma]
            data = numpy.lib.recfunctions.unstructured_to_structured(self.df[[label,label_sigma]].to_numpy(),
                                                                     numpy.dtype([("value", numpy.float32), ("sigma", numpy.float32)]))
        elif data is None:
            data = self.df[label]
            
        if numpy.iscomplexobj(data):
            asutype = gemmi.ComplexAsuData
        elif issubclass(data.dtype.type, numpy.integer):
            asutype = gemmi.IntAsuData
        elif label_sigma is not None:
            asutype = gemmi.ValueSigmaAsuData
        else:
            asutype = gemmi.FloatAsuData
        
        return asutype(self.cell, self.sg,
                       self.miller_array(), data)
    # as_asu_data()

    def fft_map(self, label=None, data=None, grid_size=None, sample_rate=3):
        if data is not None: data = data.astype(numpy.complex64) # we may want to keep complex128?
        asu = self.as_asu_data(label=label, data=data)
        if grid_size is None:
            ma = asu.transform_f_phi_to_map(sample_rate=sample_rate, exact_size=(0, 0, 0)) # half_l=True
        else:
            ma = gemmi.transform_f_phi_grid_to_map(asu.get_f_phi_on_grid(grid_size)) # half_l=False
            
        return ma
    # fft_map()

    def d_eff(self, label):
        # Effective resolution definied using FSC
        fsc = self.binned_df[label]
        a = 0.
        for i_bin, idxes in self.binned():
            a += len(idxes) * fsc[i_bin]

        fac = (a/len(self.df.index))**(1/3.)
        d_min = self.d_min_max()[0]
        ret = d_min/fac
        return ret
    # d_eff()

    def hard_sphere_kernel(self, r_ang, grid_size):
        s = 1. / self.d_spacings()
        t = 2 * numpy.pi * s * r_ang
        F_kernel = 3. * (-t * numpy.cos(t) + numpy.sin(t)) / t**3
        knl = self.fft_map(data=F_kernel, grid_size=grid_size)
        knl.array[:] += 1. / knl.unit_cell.volume # F000
        knl.array[:] /= numpy.sum(knl.array)
        return knl
    # hard_sphere_kernel()

    def scale_k_and_b(self, lab_ref, lab_scaled):
        logger.writeln("Determining k, B scales between {} and {}".format(lab_ref, lab_scaled))
        s2 = 1/self.d_spacings().to_numpy()**2
        # determine scales that minimize (|f1|-|f2|*k*e^(-b*s2/4))^2
        f1 = self.df[lab_ref].to_numpy()
        f2 = self.df[lab_scaled].to_numpy()
        if numpy.iscomplexobj(f1): f1 = numpy.abs(f1)
        if numpy.iscomplexobj(f2): f2 = numpy.abs(f2)

        sel_pos = numpy.logical_and(f1 > 0, f2 > 0) # this filters nan as well
        f1p, f2p, s2p = f1[sel_pos], f2[sel_pos], s2[sel_pos]

        # 1st step: minimize (log(|f1|)-log(|f2|*e^k*e^(-b*s2/4)))^2 starting with k=1, b=0.
        tmp = numpy.log(f2p) - numpy.log(f1p)
        # g = [dT/dk, dT/db]
        g = numpy.array([2 * numpy.sum(tmp), -numpy.sum(tmp*s2p)/2])
        H = numpy.zeros((2,2))
        H[0,0] = 2*len(f1p)
        H[1,1] = numpy.sum(s2p**2/8)
        H[0,1] = H[1,0] = -numpy.sum(s2p)/2
        x = -numpy.dot(numpy.linalg.inv(H), g)
        k1 = numpy.exp(x[0])
        B1 = x[1]
        logger.writeln(" initial estimate using log: k= {:.2e} B= {:.2e}".format(k1, B1))
        f2tmp = f2 * k1 * numpy.exp(-B1*s2/4)
        r_step0 = r_factor(f1, f2)
        r_step1 = r_factor(f1, f2tmp)
        logger.writeln(" R= {:.4f} (was: {:.4f})".format(r_step1, r_step0))

        # 2nd step: - minimize (|f1|-|f2|*k*e^(-b*s2/4))^2 iteratively (TODO with regularisation)

        def grad2(x):
            t = numpy.exp(-x[1]*s2/4)
            tmp = (f1-f2*x[0]*t)*f2*t
            return numpy.array([-2.*numpy.nansum(tmp),
                                0.5*x[0]*numpy.nansum(tmp*s2)])

        def hess2(x):
            h = numpy.zeros((2, 2))
            t = numpy.exp(-x[1]*s2/4)
            t2 = t**2
            h[0,0] = numpy.nansum(f2**2 * t2) * 2
            h[1,1] = numpy.nansum(f2 * s2**2/4 * (-f1/2*t + f2*x[0]*t2)) * x[0]
            h[1,0] = numpy.nansum(f2 * s2 * (f1/2*t - f2*x[0]*t2))
            h[0,1] = h[1,0]
            return h

        res = scipy.optimize.minimize(fun=lambda x: numpy.nansum((f1-f2*x[0]*numpy.exp(-x[1]*s2/4))**2),
                                      jac=grad2,
                                      hess=hess2,
                                      method="Newton-CG",
                                      x0=numpy.array([k1, B1]),
                                      )
        logger.writeln(str(res))
        k2, B2 = res.x
        f2tmp2 = f2 * k2 * numpy.exp(-B2*s2/4)
        r_step2 = r_factor(f1, f2tmp2)
        logger.writeln(" Least-square estimate: k= {:.2e} B= {:.2e}".format(k2, B2))
        logger.writeln(" R= {:.4f}".format(r_step2))

        if 0:
            self.setup_binning(40)        
            x = []
            y0,y1,y2,y3=[],[],[],[]
            for i_bin, idxes in self.binned():
                bin_d_min = hkldata.binned_df.d_min[i_bin]
                bin_d_max = hkldata.binned_df.d_max[i_bin]
                x.append(1/bin_d_min**2)
                y0.append(numpy.average(f1[idxes]))
                y1.append(numpy.average(f2[idxes]))
                y2.append(numpy.average(f2tmp[idxes]))
                y3.append(numpy.average(f2tmp2[idxes]))

            import matplotlib.pyplot as plt
            plt.plot(x, y0, label="FC")
            plt.plot(x, y1, label="FP")
            plt.plot(x, y2, label="FP,scaled")
            plt.plot(x, y3, label="FP,scaled2")
            plt.legend()
            plt.show()

        if r_step2 < r_step1:
            return k2, B2
        else:
            return k1, B1
    # scale_k_and_b()

    def translation_factor(self, shift):
        if type(shift) != gemmi.Position:
            shift = gemmi.Position(*shift)
        return numpy.exp(2.j*numpy.pi*numpy.dot(self.miller_array(),
                                                self.cell.fractionalize(shift).tolist()))
    # translation_factor()
    def translate(self, lab, shift):
        # apply phase shift
        assert numpy.iscomplexobj(self.df[lab])
        self.df[lab] *= self.translation_factor(shift)
    # translate()

    def write_mtz(self, mtz_out, labs, types=None, phase_label_decorator=None,
                  exclude_000=True):
        if types is None: types = {}
        if exclude_000:
            df = self.df.query("H!=0 | K!=0 | L!=0")
        else:
            df = self.df
            
        ndata = sum(2 if numpy.iscomplexobj(df[lab]) else 1 for lab in labs)

        data = numpy.empty((len(df.index), ndata + 3), dtype=numpy.float32)
        data[:,:3] = df[["H","K","L"]]
        idx = 3
        for lab in labs:
            if numpy.iscomplexobj(df[lab]):
                data[:,idx] = numpy.abs(df[lab])
                data[:,idx+1] = numpy.angle(df[lab], deg=True)
                idx += 2
            else:
                data[:,idx] = df[lab]
                idx += 1

        mtz = gemmi.Mtz()
        mtz.spacegroup = self.sg
        mtz.cell = self.cell
        mtz.add_dataset('HKL_base')
        for label in ['H', 'K', 'L']: mtz.add_column(label, 'H')

        for lab in labs:
            if numpy.iscomplexobj(df[lab]):
                mtz.add_column(lab, "F")
                if phase_label_decorator is None:
                    plab = ("PH"+lab).replace("FWT", "WT")
                else:
                    plab = phase_label_decorator(lab)
                mtz.add_column(plab, "P")
            else:
                typ = types.get(lab)
                if typ is None:
                    if issubclass(df[lab].dtype.type, numpy.integer):
                        typ = "I"
                    else:
                        typ = "R"
                mtz.add_column(lab, typ)
    
        mtz.set_data(data)
        mtz.write_to_file(mtz_out)
    # write_mtz()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import argparse
import os
import gemmi
import numpy
import json
import servalcat # for version
from servalcat.utils import logger
from servalcat import utils
from servalcat.refine.refine import Geom, Refine

def add_arguments(parser):
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument('--model',
                        help='Input atomic model file')
    group.add_argument('--update_dictionary', 
                       help="Dictionary file to be updated")
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    parser.add_argument('--ligand', nargs="*", action="append",
                        help="restraint dictionary cif file(s)")
    parser.add_argument('--ncycle', type=int, default=10,
                        help="number of CG cycles (default: %(default)d)")
    parser.add_argument('--hydrogen', default="all", choices=["all", "yes", "no", "full"],
                        help="all: add riding hydrogen atoms, yes: use hydrogen atoms if present, no: remove hydrogen atoms in input. "
                        "Default: %(default)s")
    parser.add_argument('--randomize', type=float, default=0,
                        help='Shake coordinates with specified rmsd')
    parser.add_argument('--keywords', nargs='+', action="append",
                        help="refmac keyword(s)")
    parser.add_argument('--keyword_file', nargs='+', action="append",
                        help="refmac keyword file(s)")
    parser.add_argument('-o','--output_prefix')

# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def main(args):
    if args.model:
        if not args.output_prefix:
            args.output_prefix = utils.fileio.splitext(os.path.basename(args.model))[0] + "_refined"
        st = utils.fileio.read_structure(args.model)
        utils.model.setup_entities(st, clear=True, force_subchain_names=True)
        st.assign_cis_flags()
        if st.ncs:
            st2 = st.clone()
            logger.writeln("Take NCS constraints into account.")
            st2.expand_ncs(gemmi.HowToNameCopiedChain.Dup)
            utils.fileio.write_model(st2, file_name="input_expanded.pdb")

        if args.ligand: args.ligand = sum(args.ligand, [])
        monlib = utils.restraints.load_monomer_library(st, monomer_dir=args.monlib, cif_files=args.ligand,
                                                       stop_for_unknowns=True)
        utils.restraints.find_and_fix_links(st, monlib) # should remove unknown id here?
        h_change = {"all":gemmi.HydrogenChange.ReAddButWater,
                    "full":gemmi.HydrogenChange.ReAdd,
                    "yes":gemmi.HydrogenChange.NoChange,
                    "no":gemmi.HydrogenChange.Remove}[args.hydrogen]
    else:
        if not args.output_prefix:
            args.output_prefix = utils.fileio.splitext(os.path.basename(args.update_dictionary))[0] + "_refined"
        doc = gemmi.cif.read(args.update_dictionary)
        for block in doc: # this block will be reused below
            st = gemmi.make_structure_from_chemcomp_block(block)
            if len(st) > 0: break
        else:
            raise SystemExit("No model in the cif file")
        if args.ligand:
            logger.writeln("WARNING: monlib and ligand are ignored in the dictionary updating mode")
            args.ligand = [args.update_dictionary]
        monlib = utils.restraints.load_monomer_library(st, monomer_dir=args.monlib, # monlib is needed for ener_lib
                                                       cif_files=[args.update_dictionary],
                                                       stop_for_unknowns=True)
        h_change = gemmi.HydrogenChange.NoChange
        
    try:
        topo = utils.restraints.prepare_topology(st, monlib, h_change=h_change,
                                                 check_hydrogen=(args.hydrogen=="yes"))
    except RuntimeError as e:
        raise SystemExit("Error: {}".format(e))

    if args.hydrogen == "full":
        for cra in st[0].all():
            if cra.atom.is_hydrogen(): cra.atom.occ = 1.

    keywords = []
    if args.keywords or args.keyword_file:
        if args.keywords: keywords = sum(args.keywords, [])
        if args.keyword_file: keywords.extend(l for f in sum(args.keyword_file, []) for l in open(f))

    geom = Geom(st, topo, monlib, shake_rms=args.randomize, refmac_keywords=keywords)
    refiner = Refine(st, geom)
    stats = refiner.run_cycles(args.ncycle)
    utils.fileio.write_model(refiner.st, args.output_prefix, pdb=True, cif=True)
    with open(args.output_prefix + "_stats.json", "w") as ofs:
        for s in stats: s["geom"] = s["geom"].to_dict()
        json.dump(stats, ofs, indent=2)
        logger.writeln("Refinement statistics saved: {}".format(ofs.name))

    if args.update_dictionary:
        # replace xyz
        pos = {cra.atom.name: cra.atom.pos.tolist() for cra in refiner.st[0].all()}
        for row in block.find("_chem_comp_atom.", ["atom_id", "x", "y", "z"]):
            p = pos[row.str(0)]
            for i in range(3):
                row[i+1] = "{:.3f}".format(p[i])
        # add description
        loop = block.find_loop("_pdbx_chem_comp_description_generator.comp_id").get_loop()
        if not loop:
            loop = block.init_loop("_pdbx_chem_comp_description_generator.", ["comp_id",
                                                                              "program_name",
                                                                              "program_version",
                                                                              "descriptor"])
        tags = [x[x.index(".")+1:] for x in loop.tags]
        row = ["" for _ in range(len(tags))]
        for tag, val in (("comp_id", st[0][0][0].name),
                         ("program_name", "servalcat"),
                         ("program_version", servalcat.__version__),
                         ("descriptor", "optimization tool")):
            if tag in tags: row[tags.index(tag)] = val
        loop.add_row(gemmi.cif.quote_list(row))
        doc.write_file(args.output_prefix + "_updated.cif", style=gemmi.cif.Style.Aligned)
# main()

if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
from servalcat.utils import logger
import os
import io
import gemmi
import string
import random
import numpy

default_proton_scale = 1.13 # scale of X-proton distance to X-H(e) distance

def decide_new_mod_id(mod_id, mods):
    # Refmac only allows up to 8 characters
    letters = string.digits + string.ascii_lowercase
    if len(mod_id) < 8:
        for l in letters:
            new_id = "{}{}{}".format(mod_id, "" if len(mod_id)==7 else "-", l)
            if new_id not in mods:
                return new_id

    # give up keeping original name
    while True: # XXX risk of infinite loop.. less likely though
        new_id = "mod" + "".join([random.choice(letters) for _ in range(4)])
        if new_id not in mods:
            return new_id
# decide_new_mod_id()

def rename_cif_modification_if_necessary(doc, known_ids):
    # FIXME Problematic if other file refers to modification that is renamed in this function - but how can we know?
    trans = {}
    for b in doc:
        for row in b.find("_chem_mod.", ["id"]):
            mod_id = row.str(0)
            if mod_id in known_ids:
                new_id = decide_new_mod_id(mod_id, known_ids)
                trans[mod_id] = new_id
                row[0] = new_id # modify id
                logger.writeln("INFO:: renaming modification id {} to {}".format(mod_id, new_id))

    # modify ids in mod_* blocks
    for mod_id in trans:
        b = doc.find_block("mod_{}".format(mod_id))
        if not b: # should raise error?
            logger.writeln("WARNING:: inconsistent mod description for {}".format(mod_id))
            continue
        b.name = "mod_{}".format(trans[mod_id]) # modify name
        for item in b:
            for tag in item.loop.tags:
                if tag.endswith(".mod_id"):
                    for row in b.find(tag[:tag.rindex(".")+1], ["mod_id"]):
                        row[0] = trans[mod_id]

    # Update mod id in links
    if trans:
        for b in doc:
            for row in b.find("_chem_link.", ["mod_id_1", "mod_id_2"]):
                for i in range(2):
                    if row.str(i) in trans:
                        row[i] = trans[row.str(i)]

    return trans
# rename_cif_modification_if_necessary()

def load_monomer_library(st, monomer_dir=None, cif_files=None, stop_for_unknowns=False,
                         ignore_monomer_dir=False):
    resnames = st[0].get_all_residue_names()

    if monomer_dir is None and not ignore_monomer_dir:
        if "CLIBD_MON" not in os.environ:
            logger.error("WARNING: CLIBD_MON is not set")
        else:
            monomer_dir = os.environ["CLIBD_MON"]

    if cif_files is None:
        cif_files = []
        
    if monomer_dir and not ignore_monomer_dir:
        if not os.path.isdir(monomer_dir):
            logger.error("ERROR: not a directory: {}".format(monomer_dir))
            return

        logger.writeln("Reading monomers from {}".format(monomer_dir))
        monlib = gemmi.read_monomer_lib(monomer_dir, resnames, ignore_missing=True)
    else:
        monlib = gemmi.MonLib()

    for f in cif_files:
        logger.writeln("Reading monomer: {}".format(f))
        doc = gemmi.cif.read(f)
        for b in doc:
            if b.find_values("_chem_comp_atom.atom_id"):
                name = b.name.replace("comp_", "")
                if name in monlib.monomers:
                    logger.writeln("WARNING:: updating monomer {} using {}".format(name, f))
                    del monlib.monomers[name]

                # Check if bond length values are included
                # This is to fail if cif file is e.g. from PDB website
                if not b.find_values("_chem_comp_bond.value_dist"):
                    raise RuntimeError("{} does not contain bond length value for {}. You need to generate restraints (e.g. using acedrg).".format(f, name))
                    
            for row in b.find("_chem_link.", ["id"]):
                link_id = row.str(0)
                if link_id in monlib.links:
                    logger.writeln("WARNING:: updating link {} using {}".format(link_id, f))
                    del monlib.links[link_id]

        # If modification id is duplicated, need to rename
        rename_cif_modification_if_necessary(doc, monlib.modifications)
        monlib.read_monomer_doc(doc)
        for b in doc:
            for row in b.find("_chem_comp.", ["id", "group"]):
                if row.str(0) in monlib.monomers:
                    monlib.monomers[row.str(0)].set_group(row.str(1))

    not_loaded = set(resnames).difference(monlib.monomers)
    if not_loaded:
        logger.writeln("WARNING: monomers not loaded: {}".format(" ".join(not_loaded)))
        
    logger.writeln("Monomer library loaded: {} monomers, {} links, {} modifications".format(len(monlib.monomers),
                                                                                          len(monlib.links),
                                                                                          len(monlib.modifications)))
    logger.writeln("       loaded monomers: {}".format(" ".join([x for x in monlib.monomers])))
    logger.writeln("")

    logger.writeln("Checking if unknown atoms exist..")

    unknown_cc = set()
    for chain in st[0]: unknown_cc.update(res.name for res in chain if res.name not in monlib.monomers)
    if unknown_cc:
        if stop_for_unknowns:
            raise RuntimeError("Provide restraint cif file(s) for {}".format(",".join(unknown_cc)))
        else:
            logger.writeln("WARNING: ad-hoc restraints will be generated for {}".format(",".join(unknown_cc)))
            logger.writeln("         it is strongly recommended to generate them using AceDRG.")
    
    return monlib
# load_monomer_library()

def prepare_topology(st, monlib, h_change, ignore_unknown_links=False, raise_error=True, check_hydrogen=False):
    # these checks can be done after sorting links
    logger.writeln("Creating restraints..")
    sio = io.StringIO()
    topo = gemmi.prepare_topology(st, monlib, h_change=h_change, warnings=sio, reorder=False,
                                  ignore_unknown_links=ignore_unknown_links)
    for l in sio.getvalue().splitlines(): logger.writeln(" " + l)
    unknown_cc = set()
    link_related = set()
    nan_hydr = set()
    for cinfo in topo.chain_infos:
        for rinfo in cinfo.res_infos:
            cc_org = monlib.monomers[rinfo.res.name] if rinfo.res.name in monlib.monomers else None
            for ia in reversed(range(len(rinfo.res))):
                atom = rinfo.res[ia]
                atom_str = "{}/{} {}/{}".format(cinfo.chain_ref.name, rinfo.res.name, rinfo.res.seqid, atom.name)
                cc = rinfo.get_final_chemcomp(atom.altloc)
                if not cc.find_atom(atom.name):
                    # warning message should have already been given by gemmi
                    if cc_org and cc_org.find_atom(atom.name):
                        if check_hydrogen or not atom.is_hydrogen():
                            link_related.add(rinfo.res.name)
                    else:
                        if check_hydrogen or not atom.is_hydrogen():
                            unknown_cc.add(rinfo.res.name)
                
                if atom.is_hydrogen() and atom.calc_flag == gemmi.CalcFlag.Dummy:
                    logger.writeln(" Warning: hydrogen {} could not be added - Check dictionary".format(atom_str))
                    unknown_cc.add(rinfo.res.name)
                elif any(numpy.isnan(atom.pos.tolist())): # TODO add NaN test before prepare_toplogy
                    logger.writeln(" Warning: {} position NaN!".format(atom_str))
                    nan_hydr.add(rinfo.res.name)

    if raise_error and (unknown_cc or link_related):
        msgs = []
        if unknown_cc: msgs.append("restraint cif file(s) for {}".format(",".join(unknown_cc)))
        if link_related: msgs.append("proper link cif file(s) for {} or check your model".format(",".join(link_related)))
        raise RuntimeError("Provide {}".format(" and ".join(msgs)))
    if raise_error and nan_hydr:
        raise RuntimeError("Some hydrogen positions became NaN. The geometry of your model may be of low quality. Consider not adding hydrogen")
    return topo
# prepare_topology()

def check_monlib_support_nucleus_distances(monlib, resnames):
    good = True
    nucl_not_found = []
    for resn in resnames:
        if resn not in monlib.monomers:
            logger.error("ERROR: monomer information of {} not loaded".format(resn))
            good = False
        else:
            mon = monlib.monomers[resn]
            no_nuc = False
            for bond in mon.rt.bonds:
                is_h = (mon.get_atom(bond.id1.atom).is_hydrogen(), mon.get_atom(bond.id2.atom).is_hydrogen())
                if any(is_h) and bond.value_nucleus != bond.value_nucleus:
                    no_nuc = True
                    break
            if no_nuc:
                nucl_not_found.append(resn)
                good = False

    if nucl_not_found:
        logger.writeln("WARNING: nucleus distance is not found for: {}".format(" ".join(nucl_not_found)))
        logger.writeln("         default scale ({}) is used for nucleus distances.".format(default_proton_scale))
    return good
# check_monlib_support_nucleus_distances()

def find_and_fix_links(st, monlib, bond_margin=1.3, find_metal_links=True, add_found=True):
    """
    Identify link ids for st.connections and find new links
    This is required for correctly recognizing link in gemmi.prepare_topology
    Note that it ignores segment IDs
    FIXME it assumes only one bond exists in a link. It may not be the case in future.
    """
    from servalcat.utils import model

    logger.writeln("Checking links defined in the model")
    for con in st.connections:
        if con.type == gemmi.ConnectionType.Hydrog: continue
        cra1, cra2 = st[0].find_cra(con.partner1, ignore_segment=True), st[0].find_cra(con.partner2, ignore_segment=True)
        if None in (cra1.atom, cra2.atom):
            logger.writeln(" WARNING: atom(s) not found for link: id= {} atom1= {} atom2= {}".format(con.link_id, con.partner1, con.partner2))
            continue
        if con.asu == gemmi.Asu.Different:
            nimage = st.cell.find_nearest_image(cra1.atom.pos, cra2.atom.pos, con.asu)
            image_idx = nimage.sym_idx
            dist = nimage.dist()
        else:
            image_idx = 0
            dist = cra1.atom.pos.dist(cra2.atom.pos)
        atoms_str = "atom1= {} atom2= {} image= {}".format(cra1, cra2, image_idx)
        if con.link_id:
            link = monlib.get_link(con.link_id)
            inv = False
            if link is None:
                logger.writeln(" WARNING: link {} not found in the library. Please provide link dictionary.".format(con.link_id))
                continue
            else:
                match, _, _ = monlib.test_link(link, cra1.residue.name, cra1.atom.name, cra2.residue.name, cra2.atom.name)
                if not match and monlib.test_link(link, cra2.residue.name, cra2.atom.name, cra1.residue.name, cra1.atom.name)[0]:
                    match = True
                    inv = True
                if not match:
                    logger.writeln(" WARNING: link id and atoms mismatch: id= {} {}".format(link.id, atoms_str))
                    continue
        else:
            link, inv, _, _ = monlib.match_link(cra1.residue, cra1.atom.name, cra1.atom.altloc,
                                                cra2.residue, cra2.atom.name, cra2.atom.altloc)
            if link:
                con.link_id = link.id
            else:
                ideal_dist = monlib.find_ideal_distance(cra1, cra2)
                logger.writeln(" Link unidentified (simple bond will be used): {} dist= {:.2f} ideal= {:.2f}".format(atoms_str,
                                                                                                                     dist,
                                                                                                                     ideal_dist))
                continue
        if link:
            logger.writeln(" Link confirmed: id= {} {} dist= {:.2f} ideal= {:.2f}".format(link.id,
                                                                                          atoms_str,
                                                                                          dist,
                                                                                          link.rt.bonds[0].value))
            if inv:
                con.partner1 = model.cra_to_atomaddress(cra2)
                con.partner2 = model.cra_to_atomaddress(cra1)
    if len(st.connections) == 0:
        logger.writeln(" no links defined in the model")

    logger.writeln("Finding new links (will {} added)".format("be" if add_found else "not be"))
    ns = gemmi.NeighborSearch(st[0], st.cell, 5.).populate()
    cs = gemmi.ContactSearch(3.1)
    cs.ignore = gemmi.ContactSearch.Ignore.AdjacentResidues # may miss polymer links not contiguous in a chain?
    results = cs.find_contacts(ns)
    onsb = set(gemmi.Element(x) for x in "ONSB")
    n_found = 0
    for r in results:
        if st.find_connection_by_cra(r.partner1, r.partner2): continue
        link, inv, _, _ = monlib.match_link(r.partner1.residue, r.partner1.atom.name, r.partner1.atom.altloc,
                                            r.partner2.residue, r.partner2.atom.name, r.partner2.atom.altloc,
                                            (r.dist / 1.4)**2)
        if inv:
            cra1, cra2 = r.partner2, r.partner1
        else:
            cra1, cra2 = r.partner1, r.partner2
        atoms_str = "atom1= {} atom2= {} image= {}".format(cra1, cra2, r.image_idx)
        if link:
            if r.dist > link.rt.bonds[0].value * bond_margin: continue
            logger.writeln(" New link found: id= {} {} dist= {:.2f} ideal= {:.2f}".format(link.id,
                                                                                          atoms_str,
                                                                                          r.dist,
                                                                                          link.rt.bonds[0].value))
        elif find_metal_links:
            # link only metal - O/N/S/B
            if r.partner1.atom.element.is_metal == r.partner2.atom.element.is_metal: continue
            if not r.partner1.atom.element in onsb and not r.partner2.atom.element in onsb: continue
            ideal_dist = monlib.find_ideal_distance(r.partner1, r.partner2)
            if r.dist > ideal_dist * bond_margin: continue
            logger.writeln(" Metal link found: {} dist= {:.2f} ideal= {:.2f}".format(atoms_str,
                                                                                     r.dist, ideal_dist))
        n_found += 1
        if not add_found: continue
        con = gemmi.Connection()
        con.name = "added{}".format(n_found)
        if link:
            con.link_id = link.id
            con.type = gemmi.ConnectionType.Covale
        else:
            con.type = gemmi.ConnectionType.MetalC
        con.asu = gemmi.Asu.Same if r.image_idx == 0 else gemmi.Asu.Different
        con.partner1 = model.cra_to_atomaddress(cra1)
        con.partner2 = model.cra_to_atomaddress(cra2)
        con.reported_distance = r.dist
        st.connections.append(con)
    if n_found == 0:
        logger.writeln(" no links found")
# find_and_fix_links()

def add_hydrogens(st, monlib, pos="elec"):
    assert pos in ("elec", "nucl")
    # perhaps this should be done outside..
    st.entities.clear()
    st.setup_entities()

    topo = prepare_topology(st, monlib, h_change=gemmi.HydrogenChange.ReAddButWater, ignore_unknown_links=True)
    
    if pos == "nucl":
        logger.writeln("Generating hydrogens at nucleus positions")
        resnames = st[0].get_all_residue_names()
        check_monlib_support_nucleus_distances(monlib, resnames)
        topo.adjust_hydrogen_distances(gemmi.Restraints.DistanceOf.Nucleus, default_scale=default_proton_scale)
    else:
        logger.writeln("Generating hydrogens at electron positions")
# add_hydrogens()



"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import gemmi
import numpy
import json
import os
import io
import shutil
import argparse
from servalcat.utils import logger
from servalcat import utils
from servalcat import spa # this is not a right style
from servalcat.spa import shift_maps
from servalcat.refmac.refmac_wrapper import prepare_crd

def add_arguments(parser):
    parser.description = 'Run REFMAC5 for SPA'

    parser.add_argument('--exe', default="refmac5",
                        help='refmac5 binary (default: %(default)s)')
    parser.add_argument("--monlib",
                        help="Monomer library path. Default: $CLIBD_MON")
    # sfcalc options
    group = parser.add_mutually_exclusive_group(required=True)
    group.add_argument("--halfmaps", nargs=2, help="Input half map files")
    group.add_argument("--map", help="Use this only if you really do not have half maps.")
    parser.add_argument('--mask',
                        help='Mask file')
    parser.add_argument('--model', required=True,
                        help='Input atomic model file')
    parser.add_argument('--mask_radius',
                        type=float, default=3,
                        help='')
    parser.add_argument('--mask_soft_edge',
                        type=float, default=0,
                        help='Add soft edge to model mask. Should use with --no_sharpen_before_mask?')
    parser.add_argument('--padding',
                        type=float, 
                        help='Default: 2*mask_radius')
    parser.add_argument('--no_mask', action='store_true')
    parser.add_argument('--invert_mask', action='store_true', help='not for refinement.')
    parser.add_argument('--pixel_size', type=float,
                        help='Override pixel size (A)')
    parser.add_argument('--resolution',
                        type=float,
                        help='')
    parser.add_argument('--no_trim',
                        action='store_true',
                        help='Keep original box (not recommended)')
    parser.add_argument('--blur',
                        type=float, default=0,
                        help='Sharpening or blurring B')
    utils.symmetry.add_symmetry_args(parser) # add --pg etc
    parser.add_argument('--contacting_only', action="store_true", help="Filter out non-contacting NCS")
    parser.add_argument('--ignore_symmetry',
                        help='Ignore symmetry information (MTRIX/_struct_ncs_oper) in the model file')
    parser.add_argument('--no_link_check', action='store_true', 
                        help='Do not find and fix link records in input model.')
    parser.add_argument("--b_before_mask", type=float,
                        help="sharpening B value for sharpen-mask-unsharpen procedure. By default it is determined automatically.")
    parser.add_argument('--no_sharpen_before_mask', action='store_true',
                        help='By default half maps are sharpened before masking by std of signal and unsharpened after masking. This option disables it.')
    parser.add_argument('--no_fix_microheterogeneity', action='store_true', 
                        help='By default it will fix microheterogeneity for Refmac')
    parser.add_argument('--no_fix_resi9999', action='store_true', 
                        help='By default it will split chain if max residue number > 9999 which is not supported by Refmac')
    parser.add_argument('--no_check_ncs_overlaps', action='store_true', 
                        help='Disable model overlap (e.g. expanded model is used with --pg) test')
    parser.add_argument('--no_check_ncs_map', action='store_true', 
                        help='Disable map NCS consistency test')
    parser.add_argument('--no_check_mask_with_model', action='store_true', 
                        help='Disable mask test using model')
    parser.add_argument("--prepare_only", action='store_true',
                        help="Stop before refinement")
    parser.add_argument("--no_refmacat", action='store_true',
                        help="By default uses gemmi for crd/rst file preparation (do not use makecif)")
    # run_refmac options
    # TODO use group! like refmac options
    parser.add_argument('--ligand', nargs="*", action="append",
                        help="restraint dictionary cif file(s)")
    parser.add_argument('--bfactor', type=float,
                        help="reset all atomic B values to specified value")
    parser.add_argument('--ncsr', default="local", choices=["local", "global"],
                        help="local or global NCS restrained (default: %(default)s)")
    parser.add_argument('--ncycle', type=int, default=10,
                        help="number of cycles in Refmac (default: %(default)d)")
    parser.add_argument('--tlscycle', type=int, default=0,
                        help="number of TLS cycles in Refmac (default: %(default)d)")
    parser.add_argument('--tlsin',
                        help="TLS parameter input for Refmac")
    parser.add_argument('--hydrogen', default="all", choices=["all", "yes", "no"],
                        help="all: add riding hydrogen atoms, yes: use hydrogen atoms if present, no: remove hydrogen atoms in input. "
                        "Default: %(default)s")
    parser.add_argument('--jellybody', action='store_true',
                        help="Use jelly body restraints")
    parser.add_argument('--jellybody_params', nargs=2, type=float,
                        metavar=("sigma", "dmax"), default=[0.01, 4.2],
                        help="Jelly body sigma and dmax (default: %(default)s)")
    parser.add_argument('--hout', action='store_true', help="write hydrogen atoms in the output model")
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--weight_auto_scale', type=float,
                       help="'weight auto' scale value. automatically determined from resolution and mask/box volume ratio if unspecified")
    parser.add_argument('--keywords', nargs='+', action="append",
                        help="refmac keyword(s)")
    parser.add_argument('--keyword_file', nargs='+', action="append",
                        help="refmac keyword file(s)")
    parser.add_argument('--external_restraints_json')
    parser.add_argument('--show_refmac_log', action='store_true',
                        help="show all Refmac log instead of summary")
    parser.add_argument('--output_prefix', default="refined",
                        help='output file name prefix (default: %(default)s)')
    parser.add_argument('--cross_validation', action='store_true',
                        help='Run cross validation')
    parser.add_argument('--cross_validation_method', default="shake", choices=["throughout", "shake"],
                        help="shake: randomize a model refined against a full map and then refine it against a half map, "
                        "throughout: use only a half map for refinement (another half map is used for error estimation) "
                        "Default: %(default)s")
    parser.add_argument('--shake_radius', default=0.3,
                        help='Shake rmsd in case of --cross_validation_method=shake (default: %(default).1f)')
    group = parser.add_mutually_exclusive_group()
    group.add_argument('--mask_for_fofc', help="Mask file for Fo-Fc map calculation")
    group.add_argument('--mask_radius_for_fofc', type=float, help="Mask radius for Fo-Fc map calculation")
    parser.add_argument('--trim_fofc_mtz', action="store_true", help="diffmap.mtz will have smaller cell (if --mask_for_fofc is given)")
    parser.add_argument("--fsc_resolution", type=float,
                        help="High resolution limit for FSC calculation. Default: Nyquist")

# add_arguments()
                        
def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def calc_fsc(st, output_prefix, maps, d_min, mask, mask_radius, soft_edge, b_before_mask, no_sharpen_before_mask, make_hydrogen, monlib,
             blur=0, d_min_fsc=None, cross_validation=False, cross_validation_method=None, st_sr=None):
    # st_sr: shaken-and-refined st in case of cross_validation_method=="shake"
    if cross_validation:
        assert len(maps) == 2
        assert cross_validation_method in ("shake", "throughout")
    if cross_validation and cross_validation_method == "shake":
        assert st_sr is not None
    else:
        assert st_sr is None
    
    logger.writeln("Calculating map-model FSC..")

    if d_min_fsc is None:
        d_min_fsc = utils.maps.nyquist_resolution(maps[0][0])
        logger.writeln("  --fsc_resolution is not specified. Using Nyquist resolution: {:.2f}".format(d_min_fsc))
    
    st = st.clone()
    if st_sr is not None: st_sr = st_sr.clone()

    if make_hydrogen == "all":
        utils.restraints.add_hydrogens(st, monlib)
        if st_sr is not None: utils.restraints.add_hydrogens(st_sr, monlib)

    if mask is not None or mask_radius is not None:
        if mask is None:
            assert maps[0][0].unit_cell == st.cell
            mask = utils.maps.mask_from_model(st, mask_radius, soft_edge=soft_edge, grid=maps[0][0])
        if no_sharpen_before_mask or len(maps) < 2:
            for ma in maps: ma[0].array[:] *= mask
        else:
            # It seems we need different B for different resolution limit
            if b_before_mask is None: b_before_mask = determine_b_before_mask(st, maps, maps[0][1], mask, d_min_fsc)
            maps = utils.maps.sharpen_mask_unsharpen(maps, mask, d_min_fsc, b=b_before_mask)
        
    hkldata = utils.maps.mask_and_fft_maps(maps, d_min_fsc)
    hkldata.df["FC"] = utils.model.calc_fc_fft(st, d_min_fsc - 1e-6, monlib=monlib, source="electron",
                                               miller_array=hkldata.miller_array())
    labs_fc = ["FC"]

    if st_sr is not None:
        hkldata.df["FC_sr"] = utils.model.calc_fc_fft(st_sr, d_min_fsc - 1e-6, monlib=monlib, source="electron",
                                                      miller_array=hkldata.miller_array())
        labs_fc.append("FC_sr")

    if blur != 0:
        logger.writeln(" Unblurring Fc with B={} for FSC calculation".format(blur))
        unblur = numpy.exp(blur/hkldata.d_spacings().to_numpy()**2/4.)
        for lab in labs_fc:
            hkldata.df[lab] *= unblur
    
    hkldata.setup_relion_binning()
    stats = spa.fsc.calc_fsc_all(hkldata, labs_fc=labs_fc, lab_f="FP",
                                 labs_half=["F_map1", "F_map2"] if len(maps)==2 else None)

    hkldata2 = hkldata.copy(d_min=d_min) # for FSCaverage at resolution for refinement # XXX more efficient way
    hkldata2.setup_relion_binning()
    stats2 = spa.fsc.calc_fsc_all(hkldata2, labs_fc=labs_fc, lab_f="FP",
                                  labs_half=["F_map1", "F_map2"] if len(maps)==2 else None)
    
    if "fsc_half" in stats:
        with numpy.errstate(invalid="ignore"): # XXX negative fsc results in nan!
            stats.loc[:,"fsc_full_sqrt"] = numpy.sqrt(2*stats.fsc_half/(1+stats.fsc_half))

    logger.writeln(stats.to_string()+"\n")

    # remove and rename columns
    for s in (stats, stats2):
        s.rename(columns=dict(fsc_FC_full="fsc_model", Rcmplx_FC_full="Rcmplx"), inplace=True)
        if cross_validation:
            if cross_validation_method == "shake":
                s.drop(columns=["fsc_FC_half1", "fsc_FC_half2", "fsc_FC_sr_full", "Rcmplx_FC_sr_full"], inplace=True)
                s.rename(columns=dict(fsc_FC_sr_half1="fsc_model_half1",
                                          fsc_FC_sr_half2="fsc_model_half2"), inplace=True)
            else:
                s.rename(columns=dict(fsc_FC_half1="fsc_model_half1",
                                          fsc_FC_half2="fsc_model_half2"), inplace=True)
        else:
            s.drop(columns=[x for x in s if x.startswith("fsc_FC") and x.endswith(("half1","half2"))], inplace=True)

    # FSCaverages
    fscavg_text  = "Map-model FSCaverages (at {:.2f} A):\n".format(d_min)
    fscavg_text += " FSCaverage(full) = {: .4f}\n".format(spa.fsc.fsc_average(stats2.ncoeffs, stats2.fsc_model))
    if cross_validation:
        fscavg_text += "Cross-validated map-model FSCaverages:\n"
        fscavg_text += " FSCaverage(half1)= {: .4f}\n".format(spa.fsc.fsc_average(stats2.ncoeffs, stats2.fsc_model_half1))
        fscavg_text += " FSCaverage(half2)= {: .4f}\n".format(spa.fsc.fsc_average(stats2.ncoeffs, stats2.fsc_model_half2))
    
    # for loggraph
    fsc_logfile = "{}_fsc.log".format(output_prefix)
    with open(fsc_logfile, "w") as ofs:
        columns = "1/resol^2 ncoef ln(Mn(|Fo|^2)) ln(Mn(|Fc|^2)) FSC(full,model) FSC_half FSC_full_sqrt FSC(half1,model) FSC(half2,model) Rcmplx(full,model)".split()
        
        ofs.write("$TABLE: Map-model FSC after refinement:\n")
        if len(maps) == 2:
            if cross_validation: fsc_cols = [5,6,7,8,9]
            else:                fsc_cols = [5,6,7]
        else: fsc_cols = [5]
        fsc_cols.append(10)
        if len(maps) == 2: ofs.write("$GRAPHS: FSC :A:1,5,6,7:\n")
        else:              ofs.write("$GRAPHS: FSC :A:1,5:\n")
        if cross_validation: ofs.write(": cross-validated FSC :A:1,8,9:\n".format(",".join(map(str,fsc_cols))))
        ofs.write(": Rcmplx :A:1,{}:\n".format(4+len(fsc_cols)))
        ofs.write(": ln(Mn(|F|^2)) :A:1,3,4:\n")
        ofs.write(": number of Fourier coeffs :A:1,2:\n")
        ofs.write("$$ {}$$\n".format(" ".join(columns[:4]+[columns[i-1] for i in fsc_cols])))
        ofs.write("$$\n")

        plot_columns = ["d_min", "ncoeffs", "power_FP", "power_FC", "fsc_model"]
        if len(maps) == 2:
            plot_columns.extend(["fsc_half", "fsc_full_sqrt"])
            if cross_validation:
                plot_columns.extend(["fsc_model_half1", "fsc_model_half2"])
        plot_columns.append("Rcmplx")
        with numpy.errstate(divide="ignore"):
            log_format = lambda x: "{:.3f}".format(numpy.log(x))
            ofs.write(stats.to_string(header=False, index=False, index_names=False, columns=plot_columns,
                                      formatters=dict(d_min=lambda x: "{:.4f}".format(1/x**2),
                                                      power_FP=log_format, power_FC=log_format)))
        ofs.write("\n")
        ofs.write("$$\n\n")
        ofs.write(fscavg_text)
    
    logger.write(fscavg_text)
    logger.writeln("Run loggraph {} to see plots.".format(fsc_logfile))
    
    # write json
    json.dump(stats.to_dict("records"),
              open("{}_fsc.json".format(output_prefix), "w"),
              indent=True)

    return fscavg_text
# calc_fsc()

def calc_fofc(st, st_expanded, maps, monlib, model_format, args):
    logger.writeln("Starting Fo-Fc calculation..")
    if not args.halfmaps: logger.writeln(" with limited functionality because half maps were not provided")
    logger.writeln(" model: {}".format(args.output_prefix+model_format))
    
    # for Fo-Fc in case of helical reconstruction, expand model more
    # XXX should we do it for FSC calculation also? Probably we should not do sharpen-unsharpen procedure for FSC calc either.
    if args.twist is not None:
        logger.writeln("Generating all helical copies in the box")
        st_expanded = st.clone()
        utils.symmetry.update_ncs_from_args(args, st_expanded, map_and_start=maps[0], filter_contacting=False)
        utils.model.expand_ncs(st_expanded)
        utils.fileio.write_model(st_expanded, args.output_prefix+"_expanded_all", pdb=True, cif=True)

    if args.mask_for_fofc:
        logger.writeln("  mask: {}".format(args.mask_for_fofc))
        mask = utils.fileio.read_ccp4_map(args.mask_for_fofc)[0]
    elif args.mask_radius_for_fofc:
        logger.writeln("  mask: using refined model with radius of {} A".format(args.mask_radius_for_fofc))
        mask = utils.maps.mask_from_model(st_expanded, args.mask_radius_for_fofc, grid=maps[0][0]) # use soft edge?
    else:
        logger.writeln("  mask: not used")
        mask = None
        
    hkldata, map_labs, stats_str = spa.fofc.calc_fofc(st_expanded, args.resolution, maps, mask=mask, monlib=monlib,
                                                      half1_only=(args.cross_validation and args.cross_validation_method == "throughout"),
                                                      sharpening_b=None if args.halfmaps else 0.) # assume already sharpened if fullmap is given
    spa.fofc.write_files(hkldata, map_labs, maps[0][1], stats_str,
                         mask=mask, output_prefix="diffmap",
                         trim_map=mask is not None, trim_mtz=args.trim_fofc_mtz)
    
    # Create Coot script
    spa.fofc.write_coot_script("{}_coot.py".format(args.output_prefix),
                               model_file="{}.pdb".format(args.output_prefix), # as Coot is not good at mmcif file..
                               mtz_file="diffmap.mtz",
                               contour_fo=None if mask is None else 1.2,
                               contour_fofc=None if mask is None else 3.0,
                               ncs_ops=st.ncs)
# calc_fofc()

def write_final_summary(st, refmac_summary, fscavg_text, output_prefix):
    if len(refmac_summary["cycles"]) > 1 and "actual_weight" in refmac_summary["cycles"][-2]:
        final_weight = refmac_summary["cycles"][-2]["actual_weight"]
    else:
        final_weight = "???"

    adpstats_txt = ""
    adp_stats = utils.model.adp_stats_per_chain(st[0])
    max_chain_len = max([len(x[0]) for x in adp_stats])
    max_num_len = max([len(str(x[1])) for x in adp_stats])
    for chain, natoms, qs in adp_stats:
        adpstats_txt += " Chain {0:{1}s}".format(chain, max_chain_len) if chain!="*" else " {0:{1}s}".format("All", max_chain_len+6)
        adpstats_txt += " ({0:{1}d} atoms) min={2:5.1f} median={3:5.1f} max={4:5.1f} A^2\n".format(natoms, max_num_len, qs[0],qs[2],qs[4])

    logger.writeln("""
=============================================================================
* Final Summary *

Rmsd from ideal
  bond lengths: {rmsbond} A
  bond  angles: {rmsangle} deg

{fscavgs}
 Run loggraph {fsclog} to see plots

ADP statistics
{adpstats}

Weight used: {final_weight}
             If you want to change the weight, give larger (looser restraints)
             or smaller (tighter) value to --weight_auto_scale=.
             
Open refined model and diffmap.mtz with COOT:
coot --script {prefix}_coot.py

List Fo-Fc map peaks in the ASU:
servalcat util map_peaks --map diffmap_normalized_fofc.mrc --model {prefix}.pdb --abs_level 4.0
=============================================================================
""".format(rmsbond=refmac_summary["cycles"][-1].get("rms_bond", "???"),
           rmsangle=refmac_summary["cycles"][-1].get("rms_angle", "???"),
           fscavgs=fscavg_text.rstrip(),
           fsclog="{}_fsc.log".format(output_prefix),
           adpstats=adpstats_txt.rstrip(),
           final_weight=final_weight,
           prefix=output_prefix))
# write_final_summary()

def lab_f_suffix(blur):
    if blur is None or blur == 0.:
        return ""
    elif blur > 0:
        return "Blur_{:.2f}".format(blur)
    else:
        return "Sharp_{:.2f}".format(-blur)
# lab_f_suffix()

def write_map_mtz(hkldata, mtz_out, map_labs, sig_lab=None, blur=0):
    nblur = 2 if blur != 0 else 1
    mean_f = hkldata.df[map_labs].abs().mean().min()
    data_labs = map_labs + ([sig_lab] if sig_lab else [])

    if mean_f < 1:
        scale = 10. / mean_f
        logger.writeln("Mean(|F|)= {:.2e} may be too small for Refmac. Applying scale= {:.1f}".format(mean_f, scale))
        for lab in data_labs:
            hkldata.df[lab] *= scale

    mtz_labs = data_labs + []
    mtz_types = {}
    if sig_lab: mtz_types[sig_lab] = "Q"

    if blur != 0:
        temp = hkldata.debye_waller_factors(b_iso=blur)
        for lab in data_labs:
            data = hkldata.df[lab]
            newlab = lab + lab_f_suffix(blur)
            if numpy.iscomplexobj(data): data = numpy.abs(data)
            hkldata.df[newlab] = data * temp
            mtz_labs.append(newlab)
            mtz_types[newlab] = "F" if lab != sig_lab else "Q"

    hkldata.write_mtz(mtz_out, labs=mtz_labs, types=mtz_types,
                      phase_label_decorator=lambda x: "P"+x[1:])
# write_map_mtz()

def determine_b_before_mask(st, maps, grid_start, mask, resolution):
    logger.writeln("Determining b_before_mask..")
    # work in masked map for the speed
    new_cell, new_shape, starts, shifts = shift_maps.determine_shape_and_shift(mask=mask,
                                                                               grid_start=grid_start,
                                                                               padding=5,
                                                                               mask_cutoff=0.5,
                                                                               noncentered=True,
                                                                               noncubic=True,
                                                                               json_out=None)
    st = st.clone()
    st.cell = new_cell
    st.spacegroup_hm = "P 1"
    for cra in st[0].all():
        cra.atom.pos += shifts
        cra.atom.b_iso = 0
        cra.atom.aniso = gemmi.SMat33f(0,0,0,0,0,0)

    newmaps = []
    for i in range(len(maps)): # Update maps
        g = gemmi.FloatGrid(maps[i][0].array*mask,
                            maps[i][0].unit_cell, maps[i][0].spacegroup)

        suba = g.get_subarray(starts, new_shape)
        new_grid = gemmi.FloatGrid(suba, new_cell, st.find_spacegroup())
        newmaps.append([new_grid]+maps[i][1:])

    hkldata = utils.maps.mask_and_fft_maps(newmaps, resolution)
    hkldata.df["FC"] = utils.model.calc_fc_fft(st, resolution - 1e-6, source="electron",
                                               miller_array=hkldata.miller_array())
    k, b = hkldata.scale_k_and_b("FC", "FP")
    return -b
# determine_b_before_mask()

def process_input(st, maps, resolution, monlib, mask_in, args,
                  shifted_model_prefix="shifted",
                  output_masked_prefix="masked_fs",
                  output_mtz_prefix="starting_map",
                  use_gemmi_prep=False, no_refmac_fix=False,
                  use_refmac=True):
    ret = {} # instructions for refinement
    maps = utils.maps.copy_maps(maps) # not to modify maps
    
    grid_start = maps[0][1]
    unit_cell = maps[0][0].unit_cell
    spacegroup = gemmi.SpaceGroup(1)
    start_xyz = numpy.array(maps[0][0].get_position(*grid_start).tolist())
    A = numpy.array(unit_cell.orthogonalization_matrix.tolist())
    center = numpy.sum(A, axis=1) / 2 #+ start_xyz

    # Create mask
    if mask_in:
        logger.writeln("Input mask file: {}".format(mask_in))
        mask = utils.fileio.read_ccp4_map(mask_in)[0]
    else:
        mask = None
    
    st.cell = unit_cell
    st.spacegroup_hm = "P 1"
    if use_refmac:
        ret["model_format"] = ".mmcif" if st.input_format == gemmi.CoorFormat.Mmcif else ".pdb"
        max_seq_num = max([max(res.seqid.num for res in chain) for model in st for chain in model])
        if max_seq_num > 9999 and ret["model_format"] == ".pdb":
            logger.writeln("Max residue number ({}) exceeds 9999. Will use mmcif format".format(max_seq_num))
            ret["model_format"] = ".mmcif"

    # workaround for Refmac
    # TODO need to check external restraints
    utils.model.setup_entities(st, clear=True, force_subchain_names=True)
    if use_refmac:
        if use_gemmi_prep:
            st.assign_cis_flags()
            h_change = {"all":gemmi.HydrogenChange.ReAddButWater,
                        "yes":gemmi.HydrogenChange.NoChange,
                        "no":gemmi.HydrogenChange.Remove}[args.hydrogen]
            topo = gemmi.prepare_topology(st, monlib, h_change=h_change, warnings=logger,
                                          reorder=True, ignore_unknown_links=False)
        elif not no_refmac_fix:
            topo = gemmi.prepare_topology(st, monlib, warnings=io.StringIO(), ignore_unknown_links=True)
        else:
            topo = None # not used
        if not no_refmac_fix:
            ret["refmac_fixes"] = utils.refmac.FixForRefmac(st, topo, 
                                                            fix_microheterogeneity=not args.no_fix_microheterogeneity and not use_gemmi_prep,
                                                            fix_resimax=not args.no_fix_resi9999,
                                                            fix_nonpolymer=False)
        chain_id_len_max = max([len(x) for x in utils.model.all_chain_ids(st)])
        if chain_id_len_max > 1 and ret["model_format"] == ".pdb":
            logger.writeln("Long chain ID (length: {}) detected. Will use mmcif format".format(chain_id_len_max))
            ret["model_format"] = ".mmcif"
        if not no_refmac_fix and ret["model_format"] == ".mmcif" and not use_gemmi_prep:
            ret["refmac_fixes"].fix_nonpolymer(st)

    if len(st.ncs) > 0 and args.ignore_symmetry:
        logger.writeln("Removing symmetry information from model.")
        st.ncs.clear()
    utils.symmetry.update_ncs_from_args(args, st, map_and_start=maps[0], filter_contacting=args.contacting_only)
    st_expanded = st.clone()
    if len(st.ncs) > 0:
        if not args.no_check_ncs_overlaps and utils.model.check_symmetry_related_model_duplication(st):
            raise SystemExit("\nError: Too many symmetery-related contacts detected.\n"
                             "It is very likely you gave symmetry-expanded model along with symmetry operators.")
        if not args.no_check_ncs_map and utils.maps.check_symmetry_related_map_values(st, maps[0][0]):
            raise SystemExit("\nError: Too small map correlation.\n"
                             "It is very likely your map does not follow symmetry of the model.")
        args.keywords.extend(utils.symmetry.ncs_ops_for_refmac(st.ncs))
        utils.model.expand_ncs(st_expanded)
        logger.writeln(" Saving expanded model: input_model_expanded.*")
        utils.fileio.write_model(st_expanded, "input_model_expanded", pdb=True, cif=True)

    if mask is not None and not args.no_check_mask_with_model:
        if not utils.maps.test_mask_with_model(mask, st_expanded):
            raise SystemExit("\nError: Model is out of mask.\n"
                             "Please check your --model and --mask. You can disable this test with --no_check_mask_with_model.")

    if mask is None and args.mask_radius:
        logger.writeln("Creating mask..")
        mask = utils.maps.mask_from_model(st_expanded, args.mask_radius, soft_edge=args.mask_soft_edge, grid=maps[0][0])
        #utils.maps.write_ccp4_map("mask_from_model.ccp4", mask)

    if use_refmac:
        logger.writeln(" Saving input model with unit cell information")
        utils.fileio.write_model(st, "starting_model", pdb=True, cif=True)
        ret["model_file"] = "starting_model" + ret["model_format"]

    if mask is not None:
        if args.invert_mask:
            logger.writeln("Inverting mask..")
            mask_max, mask_min = numpy.max(mask), numpy.min(mask)
            logger.writeln("  mask_max, mask_min= {}, {}".format(mask_max, mask_min))
            mask = mask_max + mask_min - mask
        
        # Mask maps
        if args.no_sharpen_before_mask or len(maps) < 2:
            logger.writeln("Applying mask..")
            for ma in maps: ma[0].array[:] *= mask
        else:
            logger.writeln("Sharpen-mask-unsharpen..")
            b_before_mask = args.b_before_mask
            if b_before_mask is None: b_before_mask = determine_b_before_mask(st, maps, grid_start, mask, resolution)                
            maps = utils.maps.sharpen_mask_unsharpen(maps, mask, resolution, b=b_before_mask)

        if not args.no_trim:
            logger.writeln(" Shifting maps and/or model..")
            if args.padding is None: args.padding = args.mask_radius * 2
            new_cell, new_shape, starts, shifts = shift_maps.determine_shape_and_shift(mask=mask,
                                                                                       grid_start=grid_start,
                                                                                       padding=args.padding,
                                                                                       mask_cutoff=0.5,
                                                                                       noncentered=True,
                                                                                       noncubic=True,
                                                                                       json_out=None)
            ret["shifts"] = shifts
            vol_mask = numpy.count_nonzero(mask.array>0.5)
            vol_map = new_shape[0] * new_shape[1] * new_shape[2]
            ret["vol_ratio"] = vol_mask / vol_map
            logger.writeln(" Vol_mask/Vol_map= {:.2e}".format(ret["vol_ratio"]))

            # Model may be built out of the box (with 'unit cell' translation symmetry)
            # It is only valid with original unit cell, but no longer valid with the new cell
            # It still would not work if model is built over multiple 'cells'.
            extra_shift = utils.model.translate_into_box(st,
                                                         origin=gemmi.Position(*start_xyz),
                                                         apply_shift=False)
            if numpy.linalg.norm(extra_shift) > 0:
                logger.writeln("Input model is out of the box. Required shift= {}".format(extra_shift))
                ret["shifts"] += gemmi.Position(*extra_shift)
                logger.writeln("Shift for model has been adjusted: {}".format(numpy.array(ret["shifts"].tolist())))
            
            st.cell = new_cell
            st.spacegroup_hm = "P 1"
            if use_refmac:
                logger.writeln(" Saving model in trimmed map..")
                utils.fileio.write_model(st, shifted_model_prefix, pdb=True, cif=True)
                ret["model_file"] = shifted_model_prefix + ret["model_format"]

            logger.writeln(" Trimming maps..")
            for i in range(len(maps)): # Update maps
                suba = maps[i][0].get_subarray(starts, new_shape)
                new_grid = gemmi.FloatGrid(suba, new_cell, spacegroup)
                maps[i][0] = new_grid

    if use_refmac and use_gemmi_prep:
        # TODO: make cispept, make link, remove unknown link id
        # TODO: cross validation?
        crdout = os.path.splitext(ret["model_file"])[0] + ".crd"
        ret["model_file"] = crdout
        ret["model_format"] = ".mmcif"
        args.keywords.append("make cr prepared")
        gemmi.setup_for_crd(st)
        doc = gemmi.prepare_refmac_crd(st, topo, monlib, h_change)
        doc.write_file(crdout, style=gemmi.cif.Style.NoBlankLines)
        logger.writeln("crd file written: {}".format(crdout))

    hkldata = utils.maps.mask_and_fft_maps(maps, resolution, None)
    hkldata.setup_relion_binning()
    if len(maps) == 2:
        map_labs = ["Fmap1", "Fmap2", "Fout"]
        ret["lab_f_half1"] = "Fmap1" + lab_f_suffix(args.blur)
        # TODO Add SIGF in case of half maps, when refmac is ready
        ret["lab_phi_half1"] = "Pmap1"
        ret["lab_f_half2"] = "Fmap2" + lab_f_suffix(args.blur)
        ret["lab_phi_half2"] = "Pmap2"
        utils.maps.calc_noise_var_from_halfmaps(hkldata)
        d_eff_full = hkldata.d_eff("FSCfull")
        logger.writeln("Effective resolution from FSCfull= {:.2f}".format(d_eff_full))
        ret["d_eff"] = d_eff_full
    else:
        map_labs = ["Fout"]
        sig_lab = None

    if use_refmac:
        if args.no_mask:
            logger.writeln("Saving unmasked maps as mtz file..")
            mtzout = output_mtz_prefix+".mtz"
        else:
            logger.writeln(" Saving masked maps as mtz file..")
            mtzout = output_masked_prefix+"_obs.mtz"

        hkldata.df.rename(columns=dict(F_map1="Fmap1", F_map2="Fmap2", FP="Fout"), inplace=True)
        if "shifts" in ret:
            for lab in map_labs: # apply phase shift
                logger.writeln("  applying phase shift for {} with translation {}".format(lab, -ret["shifts"]))
                hkldata.translate(lab, -ret["shifts"])

        write_map_mtz(hkldata, mtzout, map_labs=map_labs, blur=args.blur)
        ret["mtz_file"] = mtzout
        ret["lab_f"] = "Fout" + lab_f_suffix(args.blur)
        ret["lab_phi"] = "Pout"
    else:
        fac = hkldata.debye_waller_factors(b_iso=args.blur)
        if "shifts" in ret: fac *= hkldata.translation_factor(-ret["shifts"])
        for lab in ("F_map1", "F_map2", "FP"):
            if lab in hkldata.df: hkldata.df[lab] *= fac
    return hkldata, ret
# process_input()

def check_args(args):
    if not os.path.exists(args.model):
        raise SystemExit("Error: --model {} does not exist.".format(args.model))

    if args.cross_validation and not args.halfmaps:
        raise SystemExit("Error: half maps are needed when --cross_validation is given")
    
    if args.mask_for_fofc and not os.path.exists(args.mask_for_fofc):
        raise SystemExit("Error: --mask_for_fofc {} does not exist".format(args.mask_for_fofc))

    if args.mask_for_fofc and args.mask_radius_for_fofc:
        raise SystemExit("Error: you cannot specify both --mask_for_fofc and --mask_radius_for_fofc")

    if args.trim_fofc_mtz and not (args.mask_for_fofc or args.mask_radius_for_fofc):
        raise SystemExit("Error: --trim_fofc_mtz is specified but --mask_for_fofc is not given")

    if args.ligand: args.ligand = sum(args.ligand, [])

    if args.keywords:
        args.keywords = sum(args.keywords, [])
    else:
        args.keywords = []

    if args.keyword_file:
        args.keyword_file = sum(args.keyword_file, [])
        for f in args.keyword_file:
            logger.writeln("Keyword file: {}".format(f))
            assert os.path.exists(f)
    else:
        args.keyword_file = []

    if (args.twist, args.rise).count(None) == 1:
        raise SystemExit("ERROR: give both helical paramters --twist and --rise")
    if args.twist is not None:
        logger.writeln("INFO: setting --contacting_only because helical symmetry is given")
        args.contacting_only = True
    if args.no_mask:
        args.mask_radius = None
        if not args.no_trim:
            logger.writeln("WARNING: setting --no_trim because --no_mask is given")
            args.no_trim = True
        if args.mask:
            logger.writeln("WARNING: Your --mask is ignored because --no_mask is given")
            args.mask = None

    #if args.mask_soft_edge > 0:
    #    logger.writeln("INFO: --mask_soft_edge={} is given. Turning off sharpen_before_mask.".format(args.mask_soft_edge))
    #    args.no_sharpen_before_mask = True

    if args.resolution is None and args.model and utils.fileio.splitext(args.model)[1].endswith("cif"):
        doc = gemmi.cif.read(args.model)
        if len(doc) != 1:
            raise SystemExit("cannot find resolution from cif. Give --resolution")
        block = doc.sole_block()
        reso_str = block.find_value("_em_3d_reconstruction.resolution")
        try:
            args.resolution = float(reso_str)
        except:
            raise SystemExit("ERROR: _em_3d_reconstruction.resolution is invalid. Give --resolution")
        logger.writeln("WARNING: --resolution not given. Using _em_3d_reconstruction.resolution = {}".format(reso_str))

    if args.resolution is None:
        raise SystemExit("ERROR: --resolution is needed.")
# check_args()

def main(args):
    check_args(args)
    use_gemmi_prep = False
    if not args.prepare_only:
        try:
            refmac_ver = utils.refmac.check_version(args.exe)
        except OSError as e:
            raise SystemExit("Error: Cannot execute {}. Check Refmac instllation or use --exe to give the location.\n{}".format(args.exe, e))
        if not args.no_refmacat and refmac_ver and refmac_ver >= (5, 8, 404):
            logger.writeln(" will use gemmi to prepare restraints")
            use_gemmi_prep = True
        else:
            logger.writeln(" will use makecif to prepare restraints")

    logger.writeln("Input model: {}".format(args.model))
    st = utils.fileio.read_structure(args.model)
    if len(st) > 1:
        logger.writeln(" Removing models 2-{}".format(len(st)))
        for i in reversed(range(1, len(st))):
            del st[i]

    try:
        monlib = utils.restraints.load_monomer_library(st, monomer_dir=args.monlib, cif_files=args.ligand, 
                                                       stop_for_unknowns=True)
    except RuntimeError as e:
        raise SystemExit("Error: {}".format(e))

    if not args.no_link_check:
        utils.restraints.find_and_fix_links(st, monlib)

    try:
        utils.restraints.prepare_topology(st.clone(), monlib, h_change=gemmi.HydrogenChange.NoChange,
                                          check_hydrogen=(args.hydrogen=="yes"))
    except RuntimeError as e:
        raise SystemExit("Error: {}".format(e))

    if args.mask_for_fofc and not args.no_check_mask_with_model:
        mask = utils.fileio.read_ccp4_map(args.mask_for_fofc)[0]
        if not utils.maps.test_mask_with_model(mask, st):
            raise SystemExit("\nError: Model is out of mask.\n"
                             "Please check your --model and --mask_for_fofc. You can disable this test with --no_check_mask_with_model.")
        del mask

    if args.halfmaps:
        maps = utils.fileio.read_halfmaps(args.halfmaps, pixel_size=args.pixel_size)
    else:
        maps = [utils.fileio.read_ccp4_map(args.map, pixel_size=args.pixel_size)]
        
    shifted_model_prefix = "shifted"
    _, file_info = process_input(st, maps, resolution=args.resolution - 1e-6, monlib=monlib,
                                 mask_in=args.mask, args=args,
                                 shifted_model_prefix=shifted_model_prefix,
                                 use_gemmi_prep=use_gemmi_prep)
    if args.prepare_only:
        logger.writeln("\n--prepare_only is given. Stopping.")
        return
        
    args.mtz = file_info["mtz_file"]
    if args.halfmaps: # FIXME if no_mask?
        args.mtz_half = [file_info["mtz_file"], file_info["mtz_file"]]
    args.lab_phi = file_info["lab_phi"]  #"Pout0"
    args.lab_f = file_info["lab_f"]
    args.lab_sigf = None
    args.model = file_info["model_file"] # refmac xyzin
    model_format = file_info["model_format"]

    if args.cross_validation and args.cross_validation_method == "throughout":
        args.lab_f = file_info["lab_f_half1"]
        args.lab_phi = file_info["lab_phi_half1"]
        # XXX args.lab_sigf?

    chain_id_lens = [len(x) for x in utils.model.all_chain_ids(st)]
    keep_chain_ids = (chain_id_lens and max(chain_id_lens) == 1) # always kept unless one-letter chain IDs

    # FIXME if mtz is given and sfcalc() not ran?

    if not args.no_trim:
        refmac_prefix = "{}_{}".format(shifted_model_prefix, args.output_prefix)
    else:
        refmac_prefix = args.output_prefix # XXX this should be different name (nomask etc?)

    # Weight auto scale
    if args.weight_auto_scale is None:
        reso = file_info["d_eff"] if "d_eff" in file_info else args.resolution
        if "vol_ratio" in file_info:
            if "d_eff" in file_info:
                rlmc = (-9.503541, 3.129882, 15.439744)
            else:
                rlmc = (-8.329418, 3.032409, 14.381907)
            logger.writeln("Estimating weight auto scale using resolution and volume ratio")
            ws = rlmc[0] + reso*rlmc[1] +file_info["vol_ratio"]*rlmc[2]
        else:
            if "d_eff" in file_info:
                rlmc = (-5.903807, 2.870723)
            else:
                rlmc = (-4.891140, 2.746791)
            logger.writeln("Estimating weight auto scale using resolution")
            ws =  rlmc[0] + args.resolution*rlmc[1]
        args.weight_auto_scale = max(0.2, min(18.0, ws))
        logger.writeln(" Will use weight auto {:.2f}".format(args.weight_auto_scale))
        
    # Run Refmac
    refmac = utils.refmac.Refmac(prefix=refmac_prefix, args=args, global_mode="spa",
                                 keep_chain_ids=keep_chain_ids)
    refmac.set_libin(args.ligand)
    try:
        refmac_summary = refmac.run_refmac()
    except RuntimeError as e:
        raise SystemExit("Error: {}".format(e))

    # Modify output
    st, cif_ref = utils.fileio.read_structure_from_pdb_and_mmcif(refmac_prefix+model_format)
    st.entities.clear()
    st.setup_entities()

    if not args.no_trim:
        st.cell = maps[0][0].unit_cell
    
    if "refmac_fixes" in file_info:
        file_info["refmac_fixes"].modify_back(st)
    utils.model.adp_analysis(st)
    utils.fileio.write_model(st, prefix=args.output_prefix,
                             pdb=True, cif=True, cif_ref=cif_ref)

    # Take care of TLS out
    if not args.no_trim: # if no_trim, there is nothing to do.
        tlsout = refmac.tlsout()
        if os.path.exists(tlsout):
            logger.writeln("Copying tlsout")
            shutil.copyfile(refmac.tlsout(), args.output_prefix+".tls")

    # Expand sym here
    st_expanded = st.clone()
    if not all(op.given for op in st.ncs):
        utils.model.expand_ncs(st_expanded)
        utils.fileio.write_model(st_expanded, args.output_prefix+"_expanded", pdb=True, cif=True,
                                 cif_ref=cif_ref)

    if args.cross_validation and args.cross_validation_method == "shake":
        logger.writeln("Cross validation is requested.")
        refmac_prefix_shaken = refmac_prefix+"_shaken_refined"
        logger.writeln("Starting refinement using half map 1 (model is shaken first)")
        logger.writeln("In this refinement, hydrogen is removed regardless of --hydrogen option")
        if use_gemmi_prep:
            xyzin = refmac_prefix + ".crd"
            prepare_crd(utils.fileio.read_structure(refmac_prefix+model_format),
                        crdout=xyzin, ligand=[refmac_prefix+model_format],
                        make={"hydr":"n"})
        else:
            xyzin = refmac_prefix + model_format
        refmac_hm1 = refmac.copy(hklin=args.mtz_half[0],
                                 xyzin=xyzin,
                                 prefix=refmac_prefix_shaken,
                                 shake=args.shake_radius,
                                 jellybody=False, # makes no sense to use jelly body after shaking
                                 hydrogen="no") # should not use hydrogen after shaking
        if args.jellybody: logger.writeln("  Turning off jelly body")
        if "lab_f_half1" in file_info:
            refmac_hm1.lab_f = file_info["lab_f_half1"]
            refmac_hm1.lab_phi = file_info["lab_phi_half1"]
            # SIGMA?
            
        try:
            refmac_hm1.run_refmac()
        except RuntimeError as e:
            raise SystemExit("Error: {}".format(e))

        if args.hydrogen != "no": # does not work properly when 'yes' - we would need to keep hydrogen in input
            logger.writeln("Cross validation: 2nd run with hydrogen")
            if use_gemmi_prep:
                xyzin = refmac_prefix_shaken + ".crd"
                prepare_crd(utils.fileio.read_structure(refmac_prefix_shaken+model_format),
                            crdout=xyzin, ligand=[refmac_prefix+model_format],
                            make={"hydr":"a"})
            else:
                xyzin = refmac_prefix_shaken + model_format
            refmac_prefix_shaken = refmac_prefix+"_shaken_refined2"
            refmac_hm1_2 = refmac_hm1.copy(xyzin=xyzin,
                                           prefix=refmac_prefix_shaken,
                                           shake=None,
                                           hydrogen="all")
            try:
                refmac_hm1_2.run_refmac()
            except RuntimeError as e:
                raise SystemExit("Error: {}".format(e))
        
        # Modify output
        st_sr, cif_ref_sr = utils.fileio.read_structure_from_pdb_and_mmcif(refmac_prefix_shaken+model_format)
        st_sr.entities.clear()
        st_sr.setup_entities()
        if not args.no_trim:
            st_sr.cell = maps[0][0].unit_cell
            
        if "refmac_fixes" in file_info:
            file_info["refmac_fixes"].modify_back(st_sr)

        utils.fileio.write_model(st_sr, prefix=args.output_prefix+"_shaken_refined",
                                 pdb=True, cif=True, cif_ref=cif_ref_sr)

        # Expand sym here
        st_sr_expanded = st_sr.clone()
        if not all(op.given for op in st_sr.ncs):
            utils.model.expand_ncs(st_sr_expanded)
            utils.fileio.write_model(st_sr_expanded, args.output_prefix+"_shaken_refined_expanded",
                                     pdb=True, cif=True, cif_ref=cif_ref_sr)
            if args.twist is not None: # as requested by a user
                st_sr_expanded_all = st_sr.clone()
                utils.symmetry.update_ncs_from_args(args, st_sr_expanded_all, map_and_start=maps[0], filter_contacting=False)
                utils.model.expand_ncs(st_sr_expanded_all)
                utils.fileio.write_model(st_sr_expanded_all, args.output_prefix+"_shaken_refined_expanded_all", pdb=True, cif=True,
                                         cif_ref=cif_ref)
    else:
        st_sr_expanded = None

    if args.mask:
        mask = utils.fileio.read_ccp4_map(args.mask)[0]
    else:
        mask = None
        
    # Calc FSC
    fscavg_text = calc_fsc(st_expanded, args.output_prefix, maps,
                           args.resolution, mask=mask, mask_radius=args.mask_radius if not args.no_mask else None,
                           soft_edge=args.mask_soft_edge,
                           b_before_mask=args.b_before_mask,
                           no_sharpen_before_mask=args.no_sharpen_before_mask,
                           make_hydrogen=args.hydrogen,
                           monlib=monlib, cross_validation=args.cross_validation,
                           blur=args.blur,
                           d_min_fsc=args.fsc_resolution,
                           cross_validation_method=args.cross_validation_method, st_sr=st_sr_expanded)

    # Calc Fo-Fc (and updated) maps
    calc_fofc(st, st_expanded, maps, monlib, model_format, args)
    
    # Final summary
    write_final_summary(st, refmac_summary, fscavg_text, args.output_prefix)
# main()
        
if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)




"""
Author: "Keitaro Yamashita, Garib N. Murshudov"
MRC Laboratory of Molecular Biology
    
This software is released under the
Mozilla Public License, version 2.0; see LICENSE.
"""
from __future__ import absolute_import, division, print_function, generators
import argparse
import gemmi
import numpy
import pandas
import itertools
import scipy.special
import scipy.optimize
from servalcat.utils import logger
from servalcat import utils

"""
DFc = sum_j D_j F_c,j
The last Fc,n is bulk solvent contribution.
"""

def add_arguments(parser):
    parser.description = 'Sigma-A parameter estimation for crystallographic data'
    parser.add_argument('--hklin', required=True,
                        help='Input MTZ file')
    parser.add_argument('--labin', required=True,
                        help='MTZ column for F,SIGF,FREE')
    parser.add_argument('--free', type=int, default=0,
                        help='flag number for test set')
    parser.add_argument('--model', required=True, nargs="+", action="append",
                        help='Input atomic model file(s)')
    parser.add_argument("-d", '--d_min', type=float)
    #parser.add_argument('--d_max', type=float)
    parser.add_argument('--nbins', type=int, default=20,
                        help="Number of bins (default: %(default)d)")
    parser.add_argument('-s', '--source', choices=["electron", "xray", "neutron"], default="xray",
                        help="Scattering factor choice (default: %(default)s)")
    parser.add_argument('--D_as_exp',  action='store_true',
                        help="estimate D through exp(x) as a positivity constraint")
    parser.add_argument('--S_as_exp',  action='store_true',
                        help="estimate variance of unexplained signal through exp(x) as a positivity constraint")
    parser.add_argument('--no_solvent',  action='store_true',
                        help="Do not consider bulk solvent contribution")
    parser.add_argument('--use_cc',  action='store_true',
                        help="Use CC(|F1|,|F2|) to CC(F1,F2) conversion to derive D and S")
    parser.add_argument('--use', choices=["all", "work", "test"], default="all",
                        help="Which reflections to be used for the parameter estimate.")
    parser.add_argument('-o','--output_prefix', default="sigmaa",
                        help='output file name prefix (default: %(default)s)')
# add_arguments()

def parse_args(arg_list):
    parser = argparse.ArgumentParser()
    add_arguments(parser)
    return parser.parse_args(arg_list)
# parse_args()

def calc_DFc(Ds, Fcs):
    DFc = sum(Ds[i] * Fcs[i] for i in range(len(Ds)))
    return DFc
# calc_DFc()

def calc_abs_DFc(Ds, Fcs):
    DFc = sum(Ds[i] * Fcs[i] for i in range(len(Ds)))
    return numpy.abs(DFc)
# calc_abs_DFc()

def deriv_DFc2_and_DFc_dDj(Ds, Fcs):
    """
    [(d/dDj |sum(Dk * Fc,k)|^2,
      d/dDj |sum(Dk * Fc,k)|), ....] for j = 0 .. N-1
    """
    
    DFc = sum(Ds[i] * Fcs[i] for i in range(len(Ds)))
    DFc_abs = numpy.abs(DFc)
    
    ret = []
    for j in range(len(Ds)):
        rsq = numpy.real(Fcs[j] * DFc.conj())
        ret.append((2 * rsq,
                    rsq / DFc_abs))
    return DFc_abs, ret
# deriv_DFc2_and_DFc_dDj()

def mlf_acentric(Fo, varFo, Fcs, Ds, S, epsilon):
    # https://doi.org/10.1107/S0907444911001314
    # eqn (4)
    Sigma = 2 * varFo + epsilon * S
    DFc = calc_abs_DFc(Ds, Fcs)
    ret = numpy.log(2) + numpy.log(Fo) - numpy.log(Sigma)
    ret += -(Fo**2 + DFc**2)/Sigma
    ret += gemmi.log_bessel_i0(2*Fo*DFc/Sigma)
    return -ret
# mlf_acentric()

def deriv_mlf_wrt_D_S_acentric(Fo, varFo, Fcs, Ds, S, epsilon):
    deriv = numpy.zeros(1+len(Ds))
    Sigma = 2 * varFo + epsilon * S
    Fo2 = Fo**2
    DFc, tmp = deriv_DFc2_and_DFc_dDj(Ds, Fcs)
    i1_i0_x = gemmi.bessel_i1_over_i0(2*Fo*DFc/Sigma) # m
    for i, (sqder, der) in enumerate(tmp):
        deriv[i] = -numpy.nansum(-sqder / Sigma + i1_i0_x * 2 * Fo * der / Sigma)
    
    deriv[-1] = -numpy.nansum((-1/Sigma + (Fo2 + DFc**2 - i1_i0_x * 2 * Fo * DFc) / Sigma**2) * epsilon)
    return deriv
# deriv_mlf_wrt_D_S_acentric()

def mlf_centric(Fo, varFo, Fcs, Ds, S, epsilon):
    # https://doi.org/10.1107/S0907444911001314
    # eqn (4)
    Sigma = varFo + epsilon * S
    DFc = calc_abs_DFc(Ds, Fcs)
    ret = 0.5 * (numpy.log(2 / numpy.pi) - numpy.log(Sigma))
    ret += -0.5 * (Fo**2 + DFc**2) / Sigma
    ret += gemmi.log_cosh(Fo * DFc / Sigma)
    return -ret
# mlf_centric()

def deriv_mlf_wrt_D_S_centric(Fo, varFo, Fcs, Ds, S, epsilon):
    deriv = numpy.zeros(1+len(Ds))
    Sigma = varFo + epsilon * S
    Fo2 = Fo**2
    DFc, tmp = deriv_DFc2_and_DFc_dDj(Ds, Fcs)
    tanh_x = numpy.tanh(Fo*DFc/Sigma)
    for i, (sqder, der) in enumerate(tmp):
        deriv[i] = -numpy.nansum(-0.5 * sqder / Sigma + tanh_x * Fo * der / Sigma)
    deriv[-1] = -numpy.nansum((-0.5 / Sigma + (0.5*(Fo2+DFc**2) - tanh_x * Fo*DFc)/Sigma**2)*epsilon)
    return deriv
# deriv_mlf_wrt_D_S_centric()

#import line_profiler
#profile = line_profiler.LineProfiler()
#import atexit
#atexit.register(profile.print_stats)
#@profile
def mlf(df, fc_labs, Ds, S, centric_sel, use):
    ret = 0.
    func = (mlf_acentric, mlf_centric)
    for c, work, test in centric_sel:
        if use == "all":
            cidxes = numpy.concatenate([work, test])
        else:
            cidxes = work if use == "work" else test
        Fcs = [df[lab].to_numpy()[cidxes] for lab in fc_labs]
        ret += numpy.nansum(func[c](df.FP.to_numpy()[cidxes], df.SIGFP.to_numpy()[cidxes]**2, Fcs, Ds, S, df.epsilon.to_numpy()[cidxes]))
    return ret
# mlf()

#@profile
def deriv_mlf_wrt_D_S(df, fc_labs, Ds, S, centric_sel, use):
    ret = []
    func = (deriv_mlf_wrt_D_S_acentric, deriv_mlf_wrt_D_S_centric)
    for c, work, test in centric_sel:
        if use == "all":
            cidxes = numpy.concatenate([work, test])
        else:
            cidxes = work if use == "work" else test
        Fcs = [df[lab].to_numpy()[cidxes] for lab in fc_labs]
        ret.append((func[c](df.FP.to_numpy()[cidxes], df.SIGFP.to_numpy()[cidxes]**2, Fcs, Ds, S, df.epsilon.to_numpy()[cidxes])))
    return sum(ret)
# deriv_mlf_wrt_D_S()

def determine_mlf_params_from_cc(hkldata, fc_labs, D_labs, centric_and_selections, use="all"):
    # theorhetical values
    cc_a = lambda cc: (numpy.pi/4*(1-cc**2)**2 * scipy.special.hyp2f1(3/2, 3/2, 1, cc**2) - numpy.pi/4) / (1-numpy.pi/4)
    cc_c = lambda cc: 2/(numpy.pi-2) * (cc**2*numpy.sqrt(1-cc**2) + cc * numpy.arctan(cc/numpy.sqrt(1-cc**2)) + (1-cc**2)**(3/2)-1)
    table_fsc = numpy.arange(0, 1, 1e-3)
    table_cc = [cc_a(table_fsc), cc_c(table_fsc)]

    for lab in D_labs: hkldata.binned_df[lab] = 1.
    hkldata.binned_df["S"] = 1.
    
    stats = hkldata.binned_df[["d_max", "d_min"]].copy()
    for i, labi in enumerate(fc_labs):
        stats["CC(FP,{})".format(labi)] = numpy.nan
    for i, labi in enumerate(fc_labs):
        for j in range(i+1, len(fc_labs)):
            labj = fc_labs[j]
            stats["CC({},{})".format(labi, labj)] = numpy.nan

    # sqrt of eps * c; c = 1 for acentrics and 2 for centrics
    inv_sqrt_c_eps = 1. / numpy.sqrt(hkldata.df.epsilon.to_numpy() * (hkldata.df.centric.to_numpy() + 1))
    for i_bin, _ in hkldata.binned():
        # assume they are all acentrics.. only correct by c
        if use == "all":
            cidxes = numpy.concatenate([sel[i] for sel in centric_and_selections[i_bin] for i in (1,2)])
        else:
            i = 1 if use == "work" else 2
            cidxes = numpy.concatenate([sel[i] for sel in centric_and_selections[i_bin]])
        valid_sel = numpy.isfinite(hkldata.df.FP.to_numpy()[cidxes])
        cidxes = cidxes[valid_sel]
        factor = inv_sqrt_c_eps[cidxes]
        Fo = hkldata.df.FP.to_numpy()[cidxes] * factor
        mean_Fo2 = numpy.mean(Fo**2)
        SigFo = hkldata.df.SIGFP.to_numpy()[cidxes]
        Fcs = [hkldata.df[lab].to_numpy()[cidxes] * factor for lab in fc_labs]
        mean_Fk2 = numpy.array([numpy.mean(numpy.abs(fk)**2) for fk in Fcs])
        
        # estimate D
        cc_fo_fj = [numpy.corrcoef(numpy.abs(fj), Fo)[1,0] for fj in Fcs]
        for i in range(len(fc_labs)): stats.loc[i_bin, "CC(FP,{})".format(fc_labs[i])] = cc_fo_fj[i]
        mat = [[numpy.sqrt(numpy.mean(numpy.abs(fk)**2)/mean_Fo2) * numpy.real(numpy.corrcoef(fk, fj)[1,0])
                for fk in Fcs]
               for fj in Fcs]
        A = [[numpy.sqrt(numpy.mean(numpy.abs(fk)**2) * numpy.mean(numpy.abs(fj)**2))/mean_Fo2 * numpy.real(numpy.corrcoef(fk, fj)[1,0])
                for fk in Fcs]
               for fj in Fcs]
        A = numpy.array([[numpy.real(numpy.corrcoef(fk, fj)[1,0]) for fk in Fcs] for fj in Fcs])
        v = numpy.interp(cc_fo_fj, table_cc[0], table_fsc)

        for i in range(len(fc_labs)):
            labi = fc_labs[i]
            for j in range(i+1, len(fc_labs)):
                labj = fc_labs[j]
                stats.loc[i_bin, "CC({},{})".format(labi, labj)] = numpy.real(numpy.corrcoef(Fcs[i], Fcs[j])[1,0])

        # test all signs, fixing first Fc positive.
        cc_max = -2
        for v_test in itertools.product(*((x, -x) for x in v[1:])):
            v_test = numpy.array((v[0],)+v_test)
            Dj_test = numpy.dot(numpy.linalg.pinv(A), v_test) * numpy.sqrt(mean_Fo2 / mean_Fk2)
            DFc_test = calc_abs_DFc(Dj_test, Fcs)
            cc_test = numpy.corrcoef(Fo, numpy.abs(DFc_test))[1,0]
            if cc_test > cc_max:
                cc_max = cc_test
                v_max = v_test
                DFc = DFc_test
                Dj = Dj_test

        for lab, D in zip(D_labs, Dj):
            hkldata.binned_df.loc[i_bin, lab] = D

        # estimate S
        mean_DFc2 = numpy.mean(DFc**2)
        est_fsc_fo_fc = numpy.interp(numpy.corrcoef(Fo, DFc)[1,0], table_cc[0], table_fsc)
        S = mean_Fo2 - 2 * numpy.sqrt(mean_Fo2 * mean_DFc2) * est_fsc_fo_fc + mean_DFc2 - numpy.mean(SigFo**2)
        hkldata.binned_df.loc[i_bin, "S"] = S

    logger.writeln("\nCC:")
    logger.writeln(stats.to_string())
    logger.writeln("\nEstimates:")
    logger.writeln(hkldata.binned_df.to_string())
# determine_mlf_params_from_cc()
    
def determine_mlf_params(hkldata, fc_labs, D_labs, centric_and_selections, D_as_exp=False, S_as_exp=False, use="all"):
    assert use in ("all", "work", "test")
    if D_as_exp:
        transD = numpy.exp # D = transD(x)
        transD_deriv = numpy.exp # dD/dx
        transD_inv = numpy.log # x = transD_inv(D)
    else:
        transD = lambda x: x
        transD_deriv = lambda x: 1
        transD_inv = lambda x: x

    if S_as_exp:
        transS = numpy.exp
        transS_deriv = numpy.exp
        transS_inv = numpy.log
    else:
        transS = lambda x: x
        transS_deriv = lambda x: 1
        transS_inv = lambda x: x
    
    # Initial values
    for lab in D_labs: hkldata.binned_df[lab] = 1.
    hkldata.binned_df["S"] = 10000.
    for i_bin, _ in hkldata.binned():
        if use == "all":
            idxes = numpy.concatenate([sel[i] for sel in centric_and_selections[i_bin] for i in (1,2)])
        else:
            i = 1 if use == "work" else 2
            idxes = numpy.concatenate([sel[i] for sel in centric_and_selections[i_bin]])
        valid_sel = numpy.isfinite(hkldata.df.FP[idxes]) # as there is no nan-safe numpy.corrcoef
        idxes = idxes[valid_sel]
        FC = numpy.abs(hkldata.df.FC.to_numpy()[idxes])
        FP = hkldata.df.FP.to_numpy()[idxes]
        D = numpy.corrcoef(FP, FC)[1,0]
        hkldata.binned_df.loc[i_bin, D_labs[0]] = D
        hkldata.binned_df.loc[i_bin, "S"] = numpy.var(FP - D * FC)

    for D_lab in D_labs:
        if hkldata.binned_df[D_lab].min() <= 0:
            min_D = hkldata.binned_df[D_lab][hkldata.binned_df[D_lab] > 0].min() * 0.1
            logger.writeln("WARNING: negative {} is detected from initial estimates. Replacing it using minimum positive value {:.2e}".format(D_lab, min_D))
            hkldata.binned_df[D_lab].where(hkldata.binned_df[D_lab] > 0, min_D, inplace=True) # arbitrary
        
    logger.writeln("Initial estimates:")
    logger.writeln(hkldata.binned_df.to_string())

    for i_bin, idxes in hkldata.binned():
        x0 = [transD_inv(hkldata.binned_df[lab][i_bin]) for lab in D_labs] + [transS_inv(hkldata.binned_df.S[i_bin])]
        def target(x):
            return mlf(hkldata.df, fc_labs, transD(x[:-1]), transS(x[-1]), centric_and_selections[i_bin], use)
        def grad(x):
            g = deriv_mlf_wrt_D_S(hkldata.df, fc_labs, transD(x[:-1]), transS(x[-1]), centric_and_selections[i_bin], use)
            g[:-1] *= transD_deriv(x[:-1])
            g[-1] *= transS_deriv(x[-1])
            return g

        # test derivative
        if 0:
            gana = grad(x0)
            e = 1e-4
            for i in range(len(x0)):
                tmp = x0.copy()
                f0 = target(tmp)
                tmp[i] += e
                fe = target(tmp)
                gnum = (fe-f0)/e
                print("DERIV:", i, gnum, gana[i], gana[i]/gnum)

        #print("Bin", i_bin)
        res = scipy.optimize.minimize(fun=target, x0=x0, jac=grad)
        #print(res)
        
        for i, lab in enumerate(D_labs):
            hkldata.binned_df.loc[i_bin, lab] = transD(res.x[i])
        hkldata.binned_df.loc[i_bin, "S"] = transS(res.x[-1])

    logger.writeln("Refined estimates:")
    logger.writeln(hkldata.binned_df.to_string())
    return D_labs
# determine_mlf_params()

def merge_models(sts): # simply merge models. no fix in chain ids etc.
    model = gemmi.Model("1")
    for st in sts:
        for m in st:
            for c in m:
                model.add_chain(c)
    return model
# merge_models()

def process_input(hklin, labin, n_bins, free, xyzins, source, d_min=None):
    assert 1 < len(labin) < 4
    assert source in ["electron", "xray", "neutron"]
    
    mtz = gemmi.read_mtz_file(hklin)
    logger.writeln("Input mtz: {}".format(hklin))
    logger.writeln("    Unit cell: {:.4f} {:.4f} {:.4f} {:.3f} {:.3f} {:.3f}".format(*mtz.cell.parameters))
    logger.writeln("  Space group: {}".format(mtz.spacegroup.hm))
    logger.writeln("")
    
    sts = [utils.fileio.read_structure(f) for f in xyzins]
    logger.writeln("From model 1:")
    logger.writeln("    Unit cell: {:.4f} {:.4f} {:.4f} {:.3f} {:.3f} {:.3f}".format(*sts[0].cell.parameters))
    logger.writeln("  Space group: {}".format(sts[0].spacegroup_hm))
    logger.writeln("")
    
    if not mtz.cell.approx(sts[0].cell, 1e-3):
        logger.writeln("Warning: unit cell mismatch between model and mtz")
        logger.writeln("         using unit cell from mtz")

    for st in sts: st.cell = mtz.cell # mtz cell is used in any case

    sg_st = sts[0].find_spacegroup() # may be None
    sg_use = mtz.spacegroup
    if mtz.spacegroup != sg_st:
        logger.writeln("Warning: space group mismatch between model and mtz")
        if sg_st and mtz.spacegroup.point_group_hm() == sg_st.point_group_hm():
            logger.writeln("         using space group from model")
            sg_use = sg_st
        else:
            logger.writeln("         using space group from mtz")
        logger.writeln("")

    for st in sts: st.spacegroup_hm = sg_use.hm
    mtz.spacegroup = sg_use

    newlabels = ["FP","SIGFP"]
    if len(labin) == 3: newlabels.append("FREE")
    hkldata = utils.hkl.hkldata_from_mtz(mtz, labin, newlabels=newlabels)
    bad_sigma = hkldata.df.SIGFP <= 0
    n_bad_sigma = bad_sigma.sum()
    if n_bad_sigma > 0:
        logger.writeln("Removing {} reflections with SIGF<=0".format(n_bad_sigma))
        hkldata.df = hkldata.df[~bad_sigma]
    hkldata.switch_to_asu()
    if 0: # intensity conversion
        hkldata.df.FP.where(hkldata.df.FP > 0, 0, inplace=True)
        hkldata.df["FP"] = numpy.sqrt(hkldata.df["FP"])
        hkldata.df["SIGFP"] /= hkldata.df["FP"] + numpy.sqrt(hkldata.df["SIGFP"] + hkldata.df["FP"]**2)

        
    if d_min is None:
        d_min = hkldata.d_min_max()[0]
    else:
        hkldata = hkldata.copy(d_min=d_min)
        
    hkldata.complete()
    hkldata.sort_by_resolution()
    hkldata.calc_epsilon()
    hkldata.calc_centric()
    hkldata.setup_binning(n_bins=n_bins)
    logger.writeln("Data completeness: {:.2f}%".format(hkldata.completeness()*100.))

    fc_labs = []
    for i, st in enumerate(sts):
        lab = "FC{}".format(i)
        hkldata.df[lab] = utils.model.calc_fc_fft(st, d_min-1e-6,
                                                  source=source, mott_bethe=(source=="electron"),
                                                  miller_array=hkldata.miller_array())
        fc_labs.append(lab)

    # Create a centric selection table for faster look up
    centric_and_selections = {}
    stats = hkldata.binned_df.copy()
    stats["n_all"] = 0
    stats["n_obs"] = 0
    if "FREE" in hkldata.df:
        stats["n_work"] = 0
        stats["n_test"] = 0
        
    for i_bin, idxes in hkldata.binned():
        centric_and_selections[i_bin] = []
        n_obs = 0
        n_work, n_test = 0, 0
        for c, g2 in hkldata.df.loc[idxes].groupby("centric", sort=False):
            valid_sel = numpy.isfinite(g2.FP)
            if "FREE" in g2:
                test_sel = (g2.FREE == free).fillna(False)
                test = g2.index[test_sel]
                work = g2.index[~test_sel]
                n_work += (valid_sel & ~test_sel).sum()
                n_test += (valid_sel & test_sel).sum()
            else:
                work = g2.index
                test = type(work)([])
            centric_and_selections[i_bin].append((c, work, test))
            n_obs += numpy.sum(valid_sel)
            
        stats.loc[i_bin, "n_obs"] = n_obs
        stats.loc[i_bin, "n_all"] = len(idxes)
        if "FREE" in hkldata.df:
            stats.loc[i_bin, "n_work"] = n_work
            stats.loc[i_bin, "n_test"] = n_test
            
    stats["completeness"] = stats["n_obs"] / stats["n_all"] * 100
    logger.writeln(stats.to_string())
    return hkldata, sts, fc_labs, centric_and_selections
# process_input()

def bulk_solvent_and_lsq_scales(hkldata, sts, fc_labs, use_solvent=True):
    scaling = gemmi.Scaling(hkldata.cell, hkldata.sg)
    scaling.use_solvent = use_solvent
    scaleto = hkldata.as_asu_data(label="FP", label_sigma="SIGFP")
    scaleto.value_array["sigma"] = 1. # I guess this would be better.
    fc_asu_total = hkldata.as_asu_data(data=hkldata.df[fc_labs].sum(axis=1).to_numpy())
    if not use_solvent:
        logger.writeln("Scaling Fc with no bulk solvent contribution")
        scaling.prepare_points(fc_asu_total, scaleto)
    else:
        logger.writeln("Calculating solvent contribution..")
        d_min = hkldata.d_min_max()[0]
        grid = gemmi.FloatGrid()
        grid.setup_from(sts[0], spacing=min(0.4, (d_min-1e-6)/2))
        masker = gemmi.SolventMasker(gemmi.AtomicRadiiSet.Cctbx)
        masker.put_mask_on_float_grid(grid, merge_models(sts))
        fmask_gr = gemmi.transform_map_to_f_phi(grid)
        hkldata.df["Fmask"] = fmask_gr.get_value_by_hkl(hkldata.miller_array())
        fmask_asu = hkldata.as_asu_data("Fmask")
        scaling.prepare_points(fc_asu_total, scaleto, fmask_asu)

    scaling.fit_isotropic_b_approximately()
    logger.writeln(" initial k,b = {:.2e} {:.2e}".format(scaling.k_overall, scaling.b_overall.u11))
    scaling.fit_parameters()
    b_aniso = scaling.b_overall
    b_iso = b_aniso.trace() / 3
    b_aniso = b_aniso.added_kI(-b_iso) # subtract isotropic contribution
    logger.writeln(" k_ov= {:.2e} B_iso= {:.2e} B_aniso= {}".format(scaling.k_overall, b_iso, b_aniso))
    k_iso = hkldata.debye_waller_factors(b_iso=b_iso)
    k_aniso = hkldata.debye_waller_factors(b_cart=b_aniso)
    hkldata.df["k_aniso"] = k_aniso # we need it later when calculating stats
    
    if use_solvent:
        fc_labs.append("Fbulk")
        solvent_scale = scaling.get_solvent_scale(0.25 / hkldata.d_spacings()**2)
        hkldata.df[fc_labs[-1]] = hkldata.df.Fmask * solvent_scale

    # Apply scales.
    #  - k_aniso^-1 is applied to FP (isotropize), 
    #    but k_aniso should be applied to FC when calculating R or CC
    #  - k_iso should be applied to FC
    hkldata.df.FP /= scaling.k_overall * k_aniso
    hkldata.df.SIGFP /= scaling.k_overall * k_aniso
    for lab in fc_labs: hkldata.df[lab] *= k_iso
    
    # total Fc
    hkldata.df["FC"] = hkldata.df[fc_labs].sum(axis=1)
# bulk_solvent_and_lsq_scales()

def calculate_maps(hkldata, centric_and_selections, fc_labs, D_labs, log_out):
    nmodels = len(fc_labs)
    ofs = open(log_out, "w")
    ofs.write("""$TABLE: Statistics :
$GRAPHS
: log(Mn(|F|^2)) and variances :A:1,7,8,9,10:
: FOM :A:1,11,12:
: D :A:1,{Dns}:
: DFc :A:1,{DFcns}:
: R-factor :A:1,13:
: CC :A:1,14:
: number of reflections :A:1,3,4:
$$
1/resol^2 bin n_a n_c d_max d_min log(Mn(|Fo|^2)) log(Mn(|Fc|^2)) log(Mn(|DFc|^2)) log(Sigma) FOM_a FOM_c R CC(|Fo|,|Fc|) {Ds} {DFcs}
$$
$$
""".format(Dns=",".join(map(str, range(15, 15+nmodels))),
           Ds=" ".join(D_labs),
           DFcns=",".join(map(str, range(15+nmodels, 15+nmodels*2))),
           DFcs=" ".join(["log(Mn(|{}{}|))".format(dl,fl) for dl,fl in zip(D_labs, fc_labs)]),
           ))
    tmpl = "{:.4f} {:3d} {:7d} {:7d} {:7.3f} {:7.3f} {:.4e} {:.4e} {:4e} {:.4f} {:.4f} "
    tmpl += "{: .4f} " * (nmodels * 2)
    tmpl += "{: .4e} {:.4f} {:.4f}\n"

    hkldata.df["FWT"] = 0j
    hkldata.df["DELFWT"] = 0j
    hkldata.df["FOM"] = numpy.nan
    hkldata.df["X"] = numpy.nan # for FOM
    for i_bin, _ in hkldata.binned():
        idxes = numpy.concatenate([sel[1] for sel in centric_and_selections[i_bin]]) # w/o missing reflections
        bin_d_min = hkldata.binned_df.d_min[i_bin]
        bin_d_max = hkldata.binned_df.d_max[i_bin]
        Ds = [max(0., hkldata.binned_df[lab][i_bin]) for lab in D_labs] # negative D is replaced with zero here
        DFcs = [numpy.log(Ds[i] * numpy.average(numpy.abs(hkldata.df[lab].to_numpy()[idxes])))
                for i, lab in enumerate(fc_labs)]
        S = hkldata.binned_df.S[i_bin]
        
        # 0: acentric 1: centric
        mean_fom = [0, 0]
        nrefs = [0, 0]
        for c, cidxes, nidxes in centric_and_selections[i_bin]:
            Fcs = [hkldata.df[lab].to_numpy()[cidxes] for lab in fc_labs]

            Fc = numpy.abs(hkldata.df.FC.to_numpy()[cidxes])
            phic = numpy.angle(hkldata.df.FC.to_numpy()[cidxes])
            expip = numpy.cos(phic) + 1j*numpy.sin(phic)
            Fo = hkldata.df.FP.to_numpy()[cidxes]
            SigFo = hkldata.df.SIGFP.to_numpy()[cidxes]
            epsilon = hkldata.df.epsilon.to_numpy()[cidxes]
            nrefs[c] = len(cidxes)

            DFc = calc_abs_DFc(Ds, Fcs)
            if c == 0:
                Sigma = 2 * SigFo**2 + epsilon * S
                X = 2 * Fo * DFc / Sigma
                m = gemmi.bessel_i1_over_i0(X)
                hkldata.df.loc[cidxes, "FWT"] = (2 * m * Fo - DFc) * expip
            else:
                Sigma = SigFo**2 + epsilon * S
                X = Fo * DFc / Sigma
                m = numpy.tanh(X)
                hkldata.df.loc[cidxes, "FWT"] = (m * Fo) * expip

            hkldata.df.loc[cidxes, "DELFWT"] = (m * Fo - DFc) * expip
            hkldata.df.loc[cidxes, "FOM"] = m
            hkldata.df.loc[cidxes, "X"] = X
            mean_fom[c] = numpy.mean(m)
            
            # Fill missing
            hkldata.df.loc[nidxes, "FWT"] = sum(Ds[i] * hkldata.df[lab].to_numpy()[nidxes] for i, lab in enumerate(fc_labs))

        k = hkldata.df.k_aniso.to_numpy()[idxes]
        Fc = hkldata.df.FC.to_numpy()[idxes] * k
        Fcs = [hkldata.df[lab].to_numpy()[idxes] * k for lab in fc_labs]
        Fo = hkldata.df.FP.to_numpy()[idxes] * k
        DFc = calc_abs_DFc(Ds, Fcs)
        r = numpy.sum(numpy.abs(numpy.abs(Fc)-Fo)) / numpy.sum(Fo)
        cc = numpy.corrcoef(numpy.abs(Fc), Fo)[1,0]
        ofs.write(tmpl.format(*(1/bin_d_min**2, i_bin, nrefs[0], nrefs[1], bin_d_max, bin_d_min,
                                numpy.log(numpy.average(numpy.abs(Fo)**2)),
                                numpy.log(numpy.average(numpy.abs(Fc)**2)),
                                numpy.log(numpy.average(DFc**2)),
                                numpy.log(S), mean_fom[0], mean_fom[1], r, cc) + tuple(Ds + DFcs)))
    ofs.close()
    logger.writeln("output log: {}".format(log_out))
# calculate_maps()

def main(args):
    if args.nbins < 1:
        raise SystemExit("--nbins must be > 0")

    hkldata, sts, fc_labs, centric_and_selections = process_input(hklin=args.hklin,
                                                                  labin=args.labin.split(","),
                                                                  n_bins=args.nbins,
                                                                  free=args.free,
                                                                  xyzins=sum(args.model, []),
                                                                  source=args.source,
                                                                  d_min=args.d_min)
    # Overall scaling & bulk solvent
    # FP/SIGFP will be scaled. Total FC will be added.
    bulk_solvent_and_lsq_scales(hkldata, sts, fc_labs, use_solvent=not args.no_solvent)    

    # Show R and CC
    fpa, fca, k = hkldata.as_numpy_arrays(["FP", "FC", "k_aniso"])
    fpa *= k
    fca = numpy.abs(fca) * k
    logger.writeln(" CC(Fo,Fc)= {:.4f}".format(numpy.corrcoef(fca, fpa)[0,1]))
    logger.writeln(" Rcryst= {:.4f}".format(utils.hkl.r_factor(fpa, fca)))

    # Estimate ML parameters
    D_labs = ["D{}".format(i) for i in range(len(fc_labs))]

    if args.use_cc:
        logger.writeln("Estimating sigma-A parameters from CC..")
        determine_mlf_params_from_cc(hkldata, fc_labs, D_labs, centric_and_selections, args.use)
    else:
        logger.writeln("Estimating sigma-A parameters using ML..")
        determine_mlf_params(hkldata, fc_labs, D_labs, centric_and_selections, args.D_as_exp, args.S_as_exp, args.use)

    # Calculate maps
    log_out = "{}.log".format(args.output_prefix)
    calculate_maps(hkldata, centric_and_selections, fc_labs, D_labs, log_out)

    # Write mtz file
    labs = ["FP", "SIGFP", "FOM", "FWT", "DELFWT", "FC"]
    if not args.no_solvent:
        labs.append("Fbulk")
        labs.append("Fmask")
    mtz_out = args.output_prefix+".mtz"
    hkldata.write_mtz(mtz_out, labs=labs, types={"FOM": "W", "FP":"F", "SIGFP":"Q"})
    logger.writeln("output mtz: {}".format(mtz_out))

    return hkldata
# main()
if __name__ == "__main__":
    import sys
    args = parse_args(sys.argv[1:])
    main(args)
